"""Integration tests for the BM08 HDF5 to XDI converter."""

import numpy

from ...bliss_globals import setup_globals
from .. import testing
from ..processors.bm08 import DemoBm08Hdf5ToXdiConverter


@testing.integration_fixture
def bm08_converter():
    converter = DemoBm08Hdf5ToXdiConverter()
    converter.enable()

    setup_globals.mca1.rois.set("OdaRoi", 500, 530)
    yield converter
    setup_globals.mca1.rois.remove("OdaRoi")

    converter.disable()


def bm08_demo(expo=0.2, npoints=10):
    for _ in range(2):
        test_bm08_kscan(expo=expo, npoints=npoints)


@testing.integration_test
def test_bm08_kscan(bm08_converter: DemoBm08Hdf5ToXdiConverter, expo=0.2, npoints=10):
    positions = numpy.linspace(1, 2, npoints)
    scan_info = {
        "scan_parameters": {"scan_type": "Kscan"},
        "scan_metadata_categories": ["scan_parameters"],
    }
    scan = setup_globals.pointscan(
        setup_globals.roby,
        positions,
        expo,
        setup_globals.diode1,
        setup_globals.diode2,
        setup_globals.mca1,
        setup_globals.mca2,
        scan_info=scan_info,
    )
    scan_number = scan.scan_info["scan_nb"]
    _assert_xdi_file(bm08_converter, scan_number)


@testing.demo_assert("Check bm08 XDI file for scan #{scan_number}")
def _assert_xdi_file(bm08_converter: DemoBm08Hdf5ToXdiConverter, scan_number=int):
    result = bm08_converter._future.result(timeout=bm08_converter.retry_timeout + 5)
    output_filename = result["output_filename"]
    print(output_filename)
    with open(output_filename, "r") as fh:
        lines = fh.readlines()
    expected_counters = "# energy diode1 diode2 OdaRoi_0 OdaRoi_1 OdaRoi_2 OdaRoi_3\n"
    assert expected_counters in lines, "\n".join(expected_counters)



"""Flint-side widget for BM32 XRR analysis plots."""

from __future__ import annotations

import logging
from typing import Any
from typing import Dict
from typing import Tuple

import numpy
from silx.gui import qt
from silx.gui.plot import Plot1D

from ..flint import capture_errors

logger = logging.getLogger(__name__)


class Bm32XrrAnalysisWidget(qt.QWidget):
    def __init__(self, parent=None) -> None:
        super().__init__(parent=parent)
        self._plots: Dict[Tuple[int, int], Plot1D] = {}
        layout = qt.QGridLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(4)
        self._header = qt.QLabel(parent=self)
        self._header.setText("BM32 XRR")
        self._header.setAlignment(qt.Qt.AlignLeft | qt.Qt.AlignVCenter)
        self._header.setStyleSheet("font-weight: 400; font-size: 18px; padding: 4px;")
        layout.addWidget(self._header, 0, 0, 1, 3)
        for row in range(3):
            for col in range(3):
                plot = Plot1D(parent=self)
                plot.setGraphGrid(True)
                plot.setMinimumHeight(200)
                layout.addWidget(plot, row + 1, col)
                self._plots[(row, col)] = plot

    @capture_errors
    def clear_data(self) -> None:
        self._header.setText("BM32 XRR")
        for plot in self._plots.values():
            plot.clear()

    @capture_errors
    def set_result_data(
        self,
        result: dict,
        sample_name: str = "",
    ) -> None:
        if not isinstance(result, dict):
            logger.warning("No plottable XRR data found in workflow result")
            self.clear_data()
            self.setWindowTitle(sample_name or "BM32 XRR analysis")
            for plot in self._plots.values():
                plot.setGraphTitle("No plottable data in workflow result")
            return
        title = sample_name or "BM32 XRR analysis"
        self._header.setText(title)
        self.setWindowTitle(title)
        _draw_panels_from_result(self._plots, result)


def _as_array(value: Any) -> numpy.ndarray:
    arr = numpy.asarray(value)
    if arr.ndim == 0:
        return numpy.asarray([arr.item()])
    return arr.ravel()


def _get_series(result: Dict[str, Any], key: str) -> numpy.ndarray | None:
    value = result.get(key)
    if value is None:
        return None
    arr = _as_array(value)
    if arr.size <= 1:
        return None
    return arr


def _add_curve(
    plot: Plot1D, x: numpy.ndarray | None, y: numpy.ndarray | None, legend: str
):
    if x is None or y is None:
        return
    n = min(x.size, y.size)
    if n <= 0:
        return
    plot.addCurve(x[:n], y[:n], legend=legend)


def _draw_panels_from_result(
    plots: Dict[Tuple[int, int], Plot1D], result: Dict[str, Any]
):
    for plot in plots.values():
        plot.clear()

    tth = _get_series(result, "tth")
    tth_bg = _get_series(result, "tth_bg")
    det = _get_series(result, "det")
    det_bg = _get_series(result, "det_bg")
    detm = _get_series(result, "detm")
    detm_bg = _get_series(result, "detm_bg")
    tthc = _get_series(result, "tthc")
    tthc_bg = _get_series(result, "tthc_bg")
    tthcr = _get_series(result, "tthcr")
    tthcr_bg = _get_series(result, "tthcr_bg")
    detmn = _get_series(result, "detmn")
    detmn_bg = _get_series(result, "detmn_bg")
    detmnp = _get_series(result, "detmnp")
    detmnp_bg = _get_series(result, "detmnp_bg")
    detmnpr = _get_series(result, "detmnpr")
    detmnpr_bg = _get_series(result, "detmnpr_bg")
    detmnpri_bg = _get_series(result, "detmnpri_bg")
    fpc = _get_series(result, "fpc")
    q = _get_series(result, "q")
    r = _get_series(result, "R")
    qraw = _get_series(result, "qraw")
    rraw = _get_series(result, "Rraw")
    z = _get_series(result, "z")
    delta_rho = _get_series(result, "delta_rho")
    delta_rho_bl = _get_series(result, "delta_rho_bl")
    delta_rho_nc = _get_series(result, "delta_rho_nc")
    baseline = _get_series(result, "baseline")
    rq4_bl = _get_series(result, "Rq4_bl")
    rq4_sm = _get_series(result, "Rq4_sm")
    phase = _get_series(result, "phase")
    phase_bl = _get_series(result, "phase_bl")
    minima = _get_series(result, "minima")

    qc = None
    rc = None
    if q is not None and r is not None:
        mask = r > 0
        if numpy.any(mask):
            qc = q[mask]
            rc = r[mask]

    p = plots[(0, 0)]
    _add_curve(p, tth, det, "XRR")
    _add_curve(p, tth_bg if tth_bg is not None else tth, det_bg, "BG")
    p.setGraphTitle("Raw Detector Signal")
    p.setGraphXLabel("2θ (°)")
    p.setGraphYLabel("Detector (counts)")
    p.setYAxisLogarithmic(True)

    p = plots[(1, 0)]
    _add_curve(p, tth, detm, "XRR")
    _add_curve(p, tth_bg if tth_bg is not None else tth, detm_bg, "BG")
    p.setGraphTitle("Monitor-Normalized Signal")
    p.setGraphXLabel("2θ (°)")
    p.setGraphYLabel("I / I₀")
    p.setYAxisLogarithmic(True)

    p = plots[(2, 0)]
    _add_curve(p, tthc, detmn, "XRR")
    _add_curve(p, tthc_bg if tthc_bg is not None else tthc, detmn_bg, "BG")
    p.setGraphTitle("Direct-Beam Normalized")
    p.setGraphXLabel("2θ (°)")
    p.setGraphYLabel("I / (I₀·I_db)")
    p.setYAxisLogarithmic(True)

    p = plots[(0, 1)]
    _add_curve(p, tthc, detmn, "XRR-no fc")
    _add_curve(p, tthc, detmnp, "XRR")
    _add_curve(p, tthc_bg if tthc_bg is not None else tthc, detmnp_bg, "BG")
    _add_curve(p, tthcr, detmnpr, "XRR o-db")
    _add_curve(p, tthcr_bg if tthcr_bg is not None else tthcr, detmnpr_bg, "BG o-db")
    _add_curve(p, tthcr, detmnpri_bg, "BG interp")
    _add_curve(p, tthc, fpc, "fpc")
    p.setGraphTitle("Footprint Correction and BG Interpolation")
    p.setGraphXLabel("2θ (°)")
    p.setGraphYLabel("I / (I₀·I_db·f_p)")
    p.setYAxisLogarithmic(True)

    p = plots[(1, 1)]
    _add_curve(p, q, r, "R")
    _add_curve(p, qc, rc, "R>0")
    p.setGraphTitle("Reflectivity R(q)")
    p.setGraphXLabel("q (Å⁻¹)")
    p.setGraphYLabel("R")
    p.setYAxisLogarithmic(True)

    p = plots[(2, 1)]
    if qraw is not None and rraw is not None:
        _add_curve(p, qraw, rraw * qraw**4, "raw")
    if qc is not None and rc is not None:
        _add_curve(p, qc, rc * qc**4, "R>0")
    p.setGraphTitle("q⁴R (Raw)")
    p.setGraphXLabel("q (Å⁻¹)")
    p.setGraphYLabel("q⁴R")

    p = plots[(0, 2)]
    if qc is not None and rc is not None:
        _add_curve(p, qc, rc * qc**4, "q^4R>0")
    _add_curve(p, q, rq4_bl, "bl")
    _add_curve(p, q, rq4_sm, "sm_bl")
    _add_curve(p, q, baseline, "baseline")
    if q is not None and rq4_sm is not None and minima is not None:
        idx = minima.astype(int)
        idx = idx[(idx >= 0) & (idx < min(q.size, rq4_sm.size))]
        if idx.size:
            p.addCurve(
                q[idx],
                rq4_sm[idx],
                legend="minima",
                symbol="o",
                linestyle="",
            )
    p.setGraphTitle("Baseline and Minima")
    p.setGraphXLabel("q (Å⁻¹)")
    p.setGraphYLabel("q⁴R")

    p = plots[(1, 2)]
    if q is not None and r is not None:
        _add_curve(p, q, numpy.sqrt(numpy.clip(r * q**4, 0, None)), "nc")
    _add_curve(p, q, phase, "phase")
    _add_curve(p, q, phase_bl, "phase+bl")
    p.setGraphTitle("Phased Signal √(q⁴R)")
    p.setGraphXLabel("q (Å⁻¹)")
    p.setGraphYLabel("√(q⁴R)")

    p = plots[(2, 2)]
    _add_curve(p, z, delta_rho_nc, "nc")
    _add_curve(p, z, delta_rho, "inverted")
    _add_curve(p, z, delta_rho_bl, "inverted bl")
    p.setGraphTitle("Electron Density Profile")
    p.setGraphXLabel("z (Å)")
    p.setGraphYLabel("Δρ (e⁻·Å⁻³)")



import numpy
from blissdata.redis_engine.store import DataStore
from ewokscore import Task
from ewoksdata.data.bliss import iter_bliss_scan_data


class StopScan(
    Task,
    input_names=[
        "filename",
        "scan_number",
        "lima_limits",
        "scalar_counter_limits",
        "redis_url",
        "response_stream_name",
    ],
    optional_input_names=[
        "scan_key",
        "retry_timeout",
        "retry_period",
    ],
):
    """Stream events while processing data. Currently only one event is send:
    when the summed data of all channels reached their limit. Other events
    can be send in the future such as a request to move a motor.
    """

    def run(self):
        sumdata = None
        event_stream = _EventStream(
            self.inputs.redis_url, self.inputs.response_stream_name
        )
        limits = {**self.inputs.lima_limits, **self.inputs.scalar_counter_limits}

        for scan_point_data_dict in iter_bliss_scan_data(
            self.inputs.filename,
            self.inputs.scan_number,
            scan_key=self.get_input_value("scan_key", None),
            lima_names=list(self.inputs.lima_limits),
            counter_names=list(self.inputs.scalar_counter_limits),
            retry_timeout=self.get_input_value("retry_timeout", None),
            retry_period=self.get_input_value("retry_period", None),
        ):
            if sumdata is None:
                sumdata = scan_point_data_dict
            else:
                for name in sumdata:
                    sumdata[name] = sumdata[name] + scan_point_data_dict[name]

            limits_reached = {
                k: numpy.nanmax(v) > limits[k] for k, v in sumdata.items()
            }
            if all(limits_reached.values()):
                event = {"type": "stop"}
                event_stream.send_event(event)
                break


class _EventStream:
    MAXLEN = 100

    def __init__(self, redis_url: str, stream_name: str):
        self._data_store = DataStore(redis_url)
        self._stream_name = stream_name

    def send_event(self, event: dict):
        self._data_store._redis.xadd(
            self._stream_name, event, maxlen=self.MAXLEN, approximate=True
        )



"""Test ewoks task for the demo stop_scan_xrpd_id31 workflow.

Integrates detector frames via PyFAI, accumulates the 1D profile and sends
a ``{'type': 'stop'}`` event on a Redis stream once the accumulated mean relative error
drops below the input ``acc_profile_threshold``.
"""

from __future__ import annotations

import logging
from contextlib import ExitStack
from pathlib import Path

import h5py
from blissdata.h5api.dynamic_hdf5 import File
from ewokscore import Task
from ewoksxrpd.tasks.integrate import _BaseIntegrateSingleGeometry

from .stop_scan import _EventStream

logger = logging.getLogger(__name__)


class IntegrateAndSendMsg(
    _BaseIntegrateSingleGeometry,
    input_names=[
        "filename",
        "scan_number",
        "detector_name",
        "acc_profile_threshold",
        "redis_url",
        "response_stream_name",
    ],
    optional_input_names=[
        "scan_key",
        "retry_timeout",
        "retry_period",
    ],
):
    def run(self):
        with ExitStack() as stack:
            self._run_in_stack(stack)

    def _run_in_stack(self, stack: ExitStack):
        import numpy as np
        from pyFAI.containers import Integrate1dResult

        with self._worker() as (worker, _):
            event_stream = _EventStream(
                self.inputs.redis_url, self.inputs.response_stream_name
            )

            detector_name = str(self.inputs.detector_name)

            # --- data iterator ------------------------------------------------
            if self.inputs.scan_key:
                logger.info("PyFAI integrate data from %r", self.inputs.scan_key)
                data_iterator = self.iter_bliss_data_from_memory(
                    self.inputs.scan_key,
                    lima_names=[detector_name],
                    retry_timeout=self.get_input_value("retry_timeout", None),
                    retry_period=self.get_input_value("retry_period", None),
                )
            else:
                logger.info(
                    "PyFAI integrate data from '%s::%d.%s'",
                    self.inputs.filename,
                    self.inputs.scan_number,
                )
                data_iterator_context = self.iter_bliss_data(
                    self.inputs.filename,
                    self.inputs.scan_number,
                    lima_names=[detector_name],
                    retry_timeout=self.get_input_value("retry_timeout", None),
                    retry_period=self.get_input_value("retry_period", None),
                )
                data_iterator = stack.enter_context(data_iterator_context)

            # --- integrate + accumulate ---------------------------------------
            acc_intensity = None
            acc_variance = None  # sum of sigma**2 across frames
            nframes = 0
            for i, scan_point_data_dict in enumerate(data_iterator):
                image = scan_point_data_dict[detector_name]
                result: Integrate1dResult = worker.process(image)

                frame_intensity = np.array(result.intensity)
                frame_sigma = (
                    np.array(result.sigma) if result.sigma is not None else None
                )

                if acc_intensity is None:
                    acc_intensity = frame_intensity.copy()
                    acc_variance = frame_sigma**2 if frame_sigma is not None else None
                else:
                    acc_intensity += frame_intensity
                    if frame_sigma is not None and acc_variance is not None:
                        acc_variance += frame_sigma**2
                nframes += 1

                if acc_variance is None:
                    logger.info(
                        "frames: %d - no sigma from worker, skipping stop criterion",
                        nframes,
                    )
                    continue

                # mean relative error over bins with positive accumulated intensity:
                #   rel_err = mean( sqrt(Σσ²) / ΣI )
                mask = acc_intensity > 0
                if not mask.any():
                    continue
                rel_err = float(
                    np.mean(np.sqrt(acc_variance[mask]) / acc_intensity[mask])
                )
                logger.info("rel_err: %.4f / frames: %d", rel_err, nframes)

                if rel_err < self.inputs.acc_profile_threshold:
                    event_stream.send_event({"type": "stop"})
                    logger.warning(
                        "Stopping scan: rel_err=%.4f < threshold=%.4f (frames: %d)",
                        rel_err,
                        self.inputs.acc_profile_threshold,
                        nframes,
                    )
                    break


class WriteSumFrame(
    Task,
    input_names=[
        "filename",
        "scan_number",
        "detector_name",
    ],
):
    """Sum all raw detector frames of a completed loopscan and write ``*_sum.h5``.

    Reads raw 2D frames from the scan HDF5 file, sums them across the frame
    axis, and writes the sum alongside the frame count to a ``*_sum.h5`` file
    next to the original scan file.
    """

    def run(self):

        filename = self.inputs.filename
        scan_nb = self.inputs.scan_number
        detector_name = self.inputs.detector_name

        ds_path = f"/{scan_nb}.1/measurement/{detector_name}"

        with File(filename, retry_timeout=10, retry_period=0.1) as root:
            _ = root[f"/{scan_nb}.1/end_time"]  # ensure the scan is finished
            frames = root[ds_path][()]  # type: ignore

        nframes = int(frames.shape[0]) if frames.ndim > 2 else 1
        sum_frame = frames.sum(axis=0) if frames.ndim > 2 else frames

        stem = Path(filename).stem
        sum_filename = str(Path(filename).parent / f"{stem}_{scan_nb}_sum.h5")
        with h5py.File(sum_filename, "w") as f:
            f.create_dataset("sum_frame", data=sum_frame)
            f.create_dataset("nframes", data=nframes)
        logger.info("Auto-stop sum written: %s (%d frames)", sum_filename, nframes)



from __future__ import annotations

import datetime
import functools
import logging
import time
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Union

import numpy
from silx.utils.retry import RetryTimeoutError

from ..bliss_globals import current_session
from ..flint.plotter import BasePlotter
from .models import XrpdFieldName
from .models import XrpdPlotInfo
from .plot_data import XrpdNXDataInfo
from .plot_data import create_plot_info
from .plot_data import delete_old_entries
from .plot_data import delete_plot_info
from .plot_data import get_plots
from .plot_data import get_plots_to_remove
from .plot_data import get_xrpd_nxdata_info
from .plots import Xrpd2dIntegrationPlot
from .plots import XrpdCurvePlot
from .plots import XrpdImagePlot
from .utils import axis_label

logger = logging.getLogger(__name__)


def _background_task(method):
    @functools.wraps(method)
    def wrapper(*args, **kw):
        try:
            return method(*args, **kw)
        except Exception as e:
            logging.error("XRPD plotter task failed (%s)", e, exc_info=True)
            raise

    return wrapper


class XrpdPlotter(BasePlotter):
    def __init__(self, number_of_scans: int = 0) -> None:
        super().__init__(number_of_scans)
        if self._HAS_BLISS:
            nb_removed_entries = delete_old_entries(current_session)
            if nb_removed_entries:
                logging.info(f"Removed {nb_removed_entries} old entries from Redis")

    def handle_workflow_result(
        self,
        future,
        timestamp: Union[str, datetime.datetime, None],
        lima_name: str,
        scan_name: str,
        output_url: Optional[str] = None,
        retry_timeout: int = 60,
        retry_period: int = 1,
    ):
        """Handle workflow results in a background task"""
        if output_url:
            func = self._handle_workflow_result_from_file
        else:
            func = self._handle_workflow_result_from_memory
        self._spawn(
            func,
            future,
            timestamp,
            scan_name,
            lima_name,
            output_url=output_url,
            retry_timeout=retry_timeout,
            retry_period=retry_period,
        )

    def replot(self, **retry_options) -> None:
        """Re-draw all plots."""
        for plot_info in get_plots():
            self._spawn(self._update_plot, plot_info, retry_options)

    def clear(self):
        """Clear all plots."""
        plot_infos = get_plots()
        lima_names = set()
        for plot_info in plot_infos:
            lima_names.add(plot_info.lima_name)
            self._remove_plot(plot_info)
        for lima_name in lima_names:
            self.clear_lima_plots(lima_name)

    @_background_task
    def _handle_workflow_result_from_memory(
        self,
        future,
        timestamp: Union[str, datetime.datetime, None],
        scan_name: str,
        lima_name: str,
        retry_timeout: Optional[int] = 600,
        **retry_options,
    ) -> None:
        retry_period = retry_options.get("retry_period")
        result = future.result(timeout=retry_timeout, interval=retry_period)

        if result["radial_units"] is None:
            return

        radial_name, radial_unit = result["radial_units"].split("_")
        radial_label = axis_label(radial_name, radial_unit)

        plot_data: Dict[XrpdFieldName, numpy.ndarray] = {
            "radial": result["radial"],
            "intensity": result["intensity"],
        }

        if result.get("azimuthal") is not None:
            plot_data["azimuthal"] = result["azimuthal"]
            azim_name, azim_unit = result["azimuthal_units"].split("_")
            azim_label = axis_label(azim_name, azim_unit)
        else:
            azim_label = None

        plot_info = create_plot_info(
            scan_name,
            lima_name,
            radial_label=radial_label,
            azim_label=azim_label,
            color=self._color_cycler.next(),
            plot_data=plot_data,
            timestamp=timestamp,
        )

        self._remove_old_plots()
        self._update_plot(plot_info, retry_options)

    @_background_task
    def _handle_workflow_result_from_file(
        self,
        future,
        timestamp: Union[str, datetime.datetime, None],
        scan_name: str,
        lima_name: str,
        output_url: str,
        **retry_options,
    ) -> None:
        retry_period = retry_options.get("retry_period")
        nxdata_info = self._load_nxdata_info(future, output_url, retry_period)

        if nxdata_info is None:
            return

        plot_info = create_plot_info(
            scan_name,
            lima_name,
            radial_label=nxdata_info.radial_label,
            azim_label=nxdata_info.azim_label,
            color=self._color_cycler.next(),
            plot_data=nxdata_info.plot_data,
            hdf5_url=nxdata_info.intensity_url,
            timestamp=timestamp,
        )
        self._remove_old_plots()

        while not future.done():
            self._update_plot(plot_info, retry_options)
            if retry_period:
                time.sleep(retry_period)
        self._update_plot(plot_info, retry_options)

    def _load_nxdata_info(
        self,
        future,
        nxdata_url: str,
        retry_period: float,
    ) -> Optional[XrpdNXDataInfo]:
        """Try to read from nxdata_url until the future is ready"""
        while not future.done():
            try:
                return get_xrpd_nxdata_info(
                    nxdata_url,
                    # Use retry_period as a short timeout to check the future
                    retry_timeout=retry_period,
                    retry_period=retry_period,
                )
            except RetryTimeoutError:
                pass

        if future.exception():
            return None

        # Give it a last chance for 10 seconds
        try:
            return get_xrpd_nxdata_info(
                nxdata_url,
                retry_timeout=10,
                retry_period=retry_period,
            )
        except RetryTimeoutError:
            return None

    def _update_plot(self, plot_info: XrpdPlotInfo, retry_options: dict) -> None:
        logger.debug(
            "Update XRPD plots for scan %r and detector %r (last scan: %s)",
            plot_info.scan_name,
            plot_info.lima_name,
            plot_info.is_last_scan,
        )
        for plot in self._get_plots(
            plot_info, exclude_single_scan_plots=not plot_info.is_last_scan
        ):
            plot.update_plot(plot_info.pk, retry_options)

    def _remove_plot(self, plot_info: XrpdPlotInfo) -> None:
        for plot in self._get_plots(plot_info):
            plot.remove_plot(plot_info.pk)
        delete_plot_info(plot_info)

    def _remove_old_plots(self):
        plot_infos = get_plots_to_remove(self._max_plots)
        for plot_info in plot_infos:
            self._remove_plot(plot_info)

    def _get_plots(
        self, plot_info: XrpdPlotInfo, exclude_single_scan_plots: bool = False
    ) -> Tuple[Xrpd2dIntegrationPlot] | Tuple[XrpdCurvePlot, XrpdImagePlot]:
        if "azimuthal" in plot_info.field_names:
            if exclude_single_scan_plots:
                return tuple()

            last_2d = super()._get_plot(
                f"2D Integrated {plot_info.lima_name} (last)", Xrpd2dIntegrationPlot
            )
            return (last_2d,)

        last_1d = super()._get_plot("Integrated (Last)", XrpdCurvePlot)
        if exclude_single_scan_plots:
            return (last_1d,)

        accum_1d = super()._get_plot(f"Integrated {plot_info.lima_name}", XrpdImagePlot)
        return (last_1d, accum_1d)

    def clear_lima_plots(self, lima_name: str) -> None:
        last_1d = super()._get_plot("Integrated (Last)", XrpdCurvePlot)
        last_1d.clear()
        accum_1d = super()._get_plot(f"Integrated {lima_name}", XrpdImagePlot)
        accum_1d.clear()
        last_2d = super()._get_plot(
            f"2D Integrated {lima_name} (last)", Xrpd2dIntegrationPlot
        )
        last_2d.clear()



"""Flint-side for Flint XRPD plots"""

import logging
from typing import List

from silx.gui.plot import Plot1D
from silx.gui.plot import Plot2D
from silx.gui.plot.items import Item
from silx.gui.plot.items import ItemChangedType
from silx.gui.plot.items.curve import Curve

from ..flint import capture_errors
from .models import XrpdPlotInfo
from .plot_data import append_image_data
from .plot_data import get_2d_integration_data
from .plot_data import get_curve_data

logger = logging.getLogger(__name__)


class XrpdCurveWidget(Plot1D):

    _DEFAULT_XLABEL = "Radial bins"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setGraphGrid(True)
        self.setGraphXLabel(self._DEFAULT_XLABEL)
        self.setGraphYLabel("Scattered intensity")
        self.sigItemAdded.connect(self._item_added)
        self.sigItemRemoved.connect(self._item_removed)

    def _item_added(self, item: Item) -> None:
        if isinstance(item, Curve):
            item.sigItemChanged.connect(self._curve_changed)
            self._update_xlabel()

    def _item_removed(self, item: Item) -> None:
        if isinstance(item, Curve):
            item.sigItemChanged.disconnect(self._curve_changed)
            self._update_xlabel()

    def _curve_changed(self, change_type: ItemChangedType) -> None:
        if change_type == ItemChangedType.VISIBLE:
            self._update_xlabel()

    def _update_xlabel(self) -> None:
        unique_xlabels = set(curve.getXLabel() for curve in self.getAllCurves())
        unique_xlabels.discard(None)

        if len(unique_xlabels) == 1:
            self.setGraphXLabel(unique_xlabels.pop())
        else:
            self.setGraphXLabel(self._DEFAULT_XLABEL)

    @capture_errors
    def remove_plot(self, plot_key: str) -> None:
        try:
            plot_info = XrpdPlotInfo.get(plot_key)
        except KeyError as ex:
            logger.debug("%r already removed: %s", plot_key, ex)
            return

        legend = plot_info.legend
        logger.debug("remove %r", legend)
        self.remove(legend=plot_info.legend)

    @capture_errors
    def update_plot(self, plot_key: str, retry_options: dict) -> None:
        x, y, plot_info = get_curve_data(plot_key, **retry_options)
        legend = plot_info.legend

        if y is None:
            logger.debug("XRPD curve plot %r: no data to plot", legend)
            return

        if len(x) != len(y):
            logger.error("XRPD curve plot %r: Nx=%d, Ny=%d", legend, len(x), len(y))
        else:
            logger.debug("XRPD curve plot %r: Nx=%d, Ny=%d", legend, len(x), len(y))
            self.addCurve(
                x,
                y,
                legend=legend,
                xlabel=plot_info.radial_label,
                color=plot_info.color,
                linestyle="-",
            )

    @capture_errors
    def get_labels(self) -> List[str]:
        return [legend.getName() for legend in self.getItems()]


class XrpdImageWidget(Plot2D):
    @capture_errors
    def remove_plot(self, plot_key: str) -> None:
        self.clear()

    @capture_errors
    def update_plot(self, plot_key: str, retry_options: dict) -> None:
        img = self.getImage(legend=plot_key)
        if img is None:
            current_data = None
        else:
            current_data = img.getData(copy=False)
        x, y, plot_info = append_image_data(plot_key, current_data, **retry_options)
        title = plot_info.legend

        if y is None:
            logger.debug("XRPD image plot %s: no data to plot", title)
            return

        origin = x[0], 0
        scale = x[1] - x[0], 1
        self.clear()
        self.setGraphTitle(title)
        logger.debug("XRPD image plot %s: %s points", title, len(y))
        self.addImage(
            y,
            legend=title,
            xlabel=plot_info.radial_label,
            ylabel="Scan points",
            origin=origin,
            scale=scale,
        )

    @capture_errors
    def get_labels(self) -> List[str]:
        return [self.getGraphTitle()]


class Xrpd2dIntegrationWidget(Plot2D):
    @capture_errors
    def remove_plot(self, plot_key: str) -> None:
        self.clear()

    @capture_errors
    def update_plot(self, plot_key: str, retry_options: dict) -> None:
        x, y, intensity, plot_info = get_2d_integration_data(plot_key, **retry_options)
        if x is None or y is None or intensity is None:
            return
        origin = x[0], y[0]
        scale = x[1] - x[0], y[1] - y[0]
        self.clear()
        title = plot_info.legend
        self.setGraphTitle(title)
        logger.debug(
            "XRPD 2D integration plot %r: Nx=%d, Ny=%d, Shape=%s",
            title,
            len(x),
            len(y),
            intensity.shape,
        )
        self.addImage(
            intensity,
            legend=title,
            xlabel=plot_info.radial_label,
            ylabel=plot_info.azim_label,
            origin=origin,
            scale=scale,
        )

    @capture_errors
    def get_labels(self) -> List[str]:
        return [self.getGraphTitle()]



from __future__ import annotations

from pathlib import Path
from typing import List
from typing import Literal
from typing import Optional
from typing import Tuple
from typing import Union

from pydantic import BaseModel
from pydantic import Field
from pydantic import field_validator

from ..resources import resource_filename


class TomoProcessorModel(BaseModel, validate_assignment=True):
    slice_reconstruction_workflow: str = Field(
        default="slice_reconstruction.json",
        description="Workflow file for slice reconstruction",
        examples=["slice_reconstruction.json"],
    )
    volume_reconstruction_workflow: str = Field(
        default="volume_reconstruction.json",
        description="Workflow file for volume reconstruction",
        examples=["volume_reconstruction.json"],
    )
    queue: Optional[str] = Field(
        default=None,
        description="Ewoks queue to submit the workflow to",
        examples=["tomo_queue"],
    )
    nabu_config_file: Optional[str] = None
    slice_index: Union[Literal["first", "middle", "last"], int] = "middle"
    phase_retrieval_method: Union[Literal["CTF", "Paganin", "None"], None] = "None"
    delta_beta: Union[str, float, int] = "100"
    offset_mm: Union[str, float, int] = "0"
    cor_algorithm: Union[
        int,
        float,
        Literal[
            "centered",
            "sliding-window",
            "growing-window",
            "composite-coarse-to-fine",
            "octave-accurate",
        ],
    ] = "sliding-window"
    show_last_slice: bool = Field(
        default=False,
        description="Display last reconstructed slice in Flint",
    )
    volume_reconstruction: bool = False
    output_format: Literal["tiff", "hdf5", "jp2", "edf", "vol"] = "hdf5"
    slice_index_range: Union[Literal["all"], Tuple[int, int], List[int]] = "all"

    @field_validator("slice_reconstruction_workflow")
    def check_slice_workflow(cls, v):
        if not isinstance(v, str):
            raise ValueError("slice_reconstruction_workflow must be a string")
        if not v.endswith(".json"):
            raise ValueError("slice_reconstruction_workflow must end with .json")
        if not Path(resource_filename("tomo", v)).exists():
            raise FileNotFoundError(f"Workflow file {v} does not exist")
        return v

    @field_validator("volume_reconstruction_workflow")
    def check_volume_workflow(cls, v):
        if not isinstance(v, str):
            raise ValueError("volume_reconstruction_workflow must be a string")
        if not v.endswith(".json"):
            raise ValueError("volume_reconstruction_workflow must end with .json")
        if not Path(resource_filename("tomo", v)).exists():
            raise FileNotFoundError(f"Workflow file {v} does not exist")
        return v

    @field_validator("queue")
    def check_queue(cls, v):
        if v is not None and not isinstance(v, str):
            raise ValueError("queue must be a string or None")
        return v

    @field_validator("nabu_config_file")
    def check_nabu_config_file(cls, v):
        if v is not None and not isinstance(v, str):
            raise ValueError("nabu_config_file must be a string or None")
        if v and not Path(v).exists():
            raise FileNotFoundError(f"Nabu config file {v} does not exist")
        return v

    @field_validator("slice_index")
    def check_slice_index(cls, v):
        if isinstance(v, str):
            if v not in ["first", "last", "middle"]:
                raise ValueError(
                    "slice_index must be 'first', 'last', 'middle' or an integer"
                )
        elif not isinstance(v, int):
            raise ValueError(
                "slice_index must be an integer or one of 'first', 'last', 'middle'"
            )
        return str(v)

    @field_validator("cor_algorithm")
    def check_cor_algorithm(cls, v):
        valid_positions = [
            "centered",
            "sliding-window",
            "global",
            "growing-window",
            "sino-coarse-to-fine",
            "composite-coarse-to-fine",
            "octave-accurate",
        ]
        if isinstance(v, str):
            if v not in valid_positions:
                raise ValueError(f"cor_algorithm must be one of {valid_positions}")
        return v

    @field_validator("phase_retrieval_method")
    def check_phase_method(cls, v):
        if v is None:
            return "None"
        elif v not in ["CTF", "Paganin", "None"]:
            raise ValueError("phase_retrieval_method must be 'CTF', 'Paganin' or None")
        return v

    @field_validator("delta_beta")
    def check_positive_delta_beta(cls, v):
        if float(v) <= 0:
            raise ValueError("delta_beta must be positive")
        return str(v)

    @field_validator("offset_mm")
    def validate_offset(cls, v):
        if v is None:
            return 0.0
        try:
            return float(v)
        except (TypeError, ValueError):
            raise ValueError("offset must be numeric")

    @field_validator("output_format")
    def check_output_format(cls, v):
        valid_formats = ["tiff", "hdf5", "jp2", "edf"]
        if v not in valid_formats:
            raise ValueError(f"output_format must be one of {valid_formats}")
        return v

    @field_validator("slice_index_range")
    def check_slice_index_range(cls, v):
        if v != "all":
            if (
                not isinstance(v, tuple)
                or not isinstance(v, list)
                or len(v) != 2
                or not all(isinstance(i, int) for i in v)
                or v[0] < 0
                or v[1] < v[0]
            ):
                raise ValueError(
                    "slice_index_range must be 'all' or a tuple of two non-negative integers (z_bottom, z_top) with z_top >= z_bottom"
                )
        return v



import os
import time
from contextlib import contextmanager
from typing import Generator
from typing import List

from ....bliss_globals import setup_globals
from ....processor import BlissScanType
from ....xrpd.plots import Xrpd2dIntegrationPlot
from ....xrpd.plots import XrpdCurvePlot
from ....xrpd.plots import XrpdImagePlot
from ... import testing

_RETRY_TIMEOUT = 20
_LIMA1_SLEEP = 10 if os.environ.get("CI") else 0


def test_xrpd_ct_with_1d_integration(
    processor, detectors, nrepeats: int = None, expo=0.2
):
    if nrepeats is None:
        nrepeats = processor.number_of_scans + 1

    with _scan_context() as scans:
        for _ in range(nrepeats):
            time.sleep(
                _LIMA1_SLEEP
            )  # Lima1: Camera not Prepared nor Ready nor Armed (state: 5)
            scan = processor.pct(expo, *detectors)
            scans.append(scan)
            # Workflow is getting data from Lima directly so we cannot start
            # the next scan until the workflow is done.
            testing.wait_workflows()

    for lima_name in processor.lima_names:
        expected_labels = [_get_scan_legend(processor, scans[-1], lima_name)]
        _assert_1d_plot_accum(processor._plotter, expected_labels, lima_name)

    nremoved = max(len(scans) - processor.number_of_scans, 0)
    expected_labels = [
        _get_scan_legend(processor, scan, lima_name)
        for scan in scans[nremoved:]
        for lima_name in processor.lima_names
    ]
    _assert_1d_plot_last(processor._plotter, expected_labels)


def test_xrpd_ct_with_2d_integration(
    processor, detectors, nrepeats: int = None, expo=0.2
):
    if nrepeats is None:
        nrepeats = processor.number_of_scans + 1

    with _scan_context() as scans:
        for _ in range(nrepeats):
            time.sleep(
                _LIMA1_SLEEP
            )  # Lima1: Camera not Prepared nor Ready nor Armed (state: 5)
            scan = processor.pct(expo, *detectors)
            scans.append(scan)
            # Workflow is getting data from Lima directly so we cannot start
            # the next scan until the workflow is done.
            testing.wait_workflows()

    for lima_name in processor.lima_names:
        expected_labels = [_get_scan_legend(processor, scans[-1], lima_name)]
        _assert_2d_plot_last(processor._plotter, expected_labels, lima_name)


def test_xrpd_scan_with_1d_integration(
    processors,
    detectors,
    nrepeats: int = None,
    expo=0.2,
    npoints=10,
    plotting_fails: bool = True,
):
    if nrepeats is None:
        nrepeats = min(p.number_of_scans for p in processors) + 1

    with _scan_context() as scans:
        for _ in range(nrepeats):
            time.sleep(
                _LIMA1_SLEEP
            )  # Lima1: Camera not Prepared nor Ready nor Armed (state: 5)
            scan = setup_globals.loopscan(npoints, expo, *detectors)
            scans.append(scan)

    testing.wait_workflows()

    for processor in processors:
        for scan in scans:
            for lima_name in processor.lima_names:
                _assert_1d_data(
                    lima_name, processor, scan, scan.scan_info["scan_nb"], npoints
                )

        for lima_name in processor.lima_names:
            expected_labels = [_get_scan_legend(processor, scans[-1], lima_name)]
            _assert_1d_plot_accum(processor._plotter, expected_labels, lima_name)

    if plotting_fails:
        expected_labels = [
            _get_scan_legend(processor, scan, lima_name)
            for scan in scans[max(len(scans) - processor.number_of_scans, 0) :]
            for processor in processors
            for lima_name in processor.lima_names
        ]
    else:
        expected_labels = []
    _assert_1d_plot_last(processor._plotter, expected_labels)


def test_xrpd_scan_with_2d_integration(
    processors, detectors, nrepeats: int = None, expo=0.2, npoints=10
):
    if nrepeats is None:
        nrepeats = min(p.number_of_scans for p in processors) + 1

    with _scan_context() as scans:
        for _ in range(nrepeats):
            time.sleep(
                _LIMA1_SLEEP
            )  # Lima1: Camera not Prepared nor Ready nor Armed (state: 5)
            scan = setup_globals.loopscan(npoints, expo, *detectors)
            scans.append(scan)

    testing.wait_workflows()

    for processor in processors:
        for scan in scans:
            for lima_name in processor.lima_names:
                _assert_2d_data(
                    lima_name, processor, scan, scan.scan_info["scan_nb"], npoints
                )

        for lima_name in processor.lima_names:
            expected_labels = [_get_scan_legend(processor, scans[-1], lima_name)]
            _assert_2d_plot_last(processor._plotter, expected_labels, lima_name)


def _assert_1d_plot_last(plotter, expected_labels):
    """The last 1D diffractogram of each scan is plotted as a curve."""
    _assert_plot(plotter, expected_labels, "Integrated (Last)", XrpdCurvePlot)


def _assert_1d_plot_accum(plotter, expected_labels, lima_name):
    """The 1D diffractograms of the last scan are plotted as an image."""
    _assert_plot(plotter, expected_labels, f"Integrated {lima_name}", XrpdImagePlot)


def _assert_2d_plot_last(plotter, expected_labels, lima_name):
    """The last 2D cake of the last scan is plotted as an image."""
    _assert_plot(
        plotter,
        expected_labels,
        f"2D Integrated {lima_name} (last)",
        Xrpd2dIntegrationPlot,
    )


@testing.demo_assert("Check {plot_id} plot for {expected_labels}")
def _assert_plot(plotter, expected_labels, plot_id, plot_cls):
    _ = plotter.wait_tasks(_RETRY_TIMEOUT)
    plot = plotter._get_plot(plot_id, plot_cls)
    labels = plot.get_labels()
    err_msg = f"{labels} instead of {expected_labels}"
    assert set(labels) == set(expected_labels), err_msg


def _get_scan_legend(processor, scan, lima_name) -> str:
    scan_name = processor._get_scan_name(scan)
    return f"{scan_name} ({lima_name})"


@testing.demo_assert("Check 1D integration data for {lima_name} of scan #{scan_number}")
def _assert_1d_data(lima_name, processor, scan, scan_number, npoints):
    data_keys = processor.get_data_keys(
        scan, lima_name, retry_timeout=_RETRY_TIMEOUT, retry_period=0.2
    )

    shapes = {
        name: processor.get_data(
            scan, name, retry_timeout=_RETRY_TIMEOUT, retry_period=0.2
        ).shape
        for name in data_keys
    }
    expected = processor._expected_data_shapes_1d(lima_name, npoints)
    assert shapes == expected, shapes


@testing.demo_assert("Check 2D integration data for {lima_name} of scan #{scan_number}")
def _assert_2d_data(lima_name, processor, scan, scan_number, npoints):
    data_keys = processor.get_data_keys(
        scan, lima_name, retry_timeout=_RETRY_TIMEOUT, retry_period=0.2
    )

    shapes = {
        name: processor.get_data(
            scan, name, retry_timeout=_RETRY_TIMEOUT, retry_period=0.2
        ).shape
        for name in data_keys
    }
    expected = {
        f"{lima_name}:chi": (360,),
        f"{lima_name}:q": (4096,),
        f"{lima_name}:intensity": (npoints, 360, 4096),
        f"{lima_name}:points": (npoints,),
    }
    assert shapes == expected, shapes


@contextmanager
def _scan_context() -> Generator[List[BlissScanType], None, None]:
    scans = []
    try:
        yield scans
    finally:
        # Holds references to processor instances
        for scan in scans:
            scan._Scan__user_scan_meta = None



from __future__ import annotations

from contextlib import contextmanager
from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional
from typing import Union

import fabio
import numpy
from ewoksutils.task_utils import task_inputs

from ..bliss_globals import setup_globals
from ..persistent.parameters import ParameterInfo
from ..streamline.scanner import StreamlineScanner
from ..utils import validators
from . import optimize_exposure
from .stop_scan_preset import StopIntegrateSum
from .utils import ensure_shutter_open


class Id31StreamlineScanner(
    StreamlineScanner,
    parameters=[
        ParameterInfo(
            "optimize_pixel_value", category="exposure/attenuator", validator=float
        ),
        ParameterInfo(
            "optimize_nb_frames", category="exposure/attenuator", validator=int
        ),
        ParameterInfo(
            "optimize_min_exposure_time",
            category="exposure/attenuator",
            validator=float,
        ),
        ParameterInfo(
            "optimize_max_exposure_time",
            category="exposure/attenuator",
            validator=float,
        ),
        ParameterInfo("default_attenuator", category="exposure/attenuator"),
        ParameterInfo("attenuator_name", category="names"),
        ParameterInfo("newflat", category="Flat-field", validator=validators.is_file),
        ParameterInfo("oldflat", category="Flat-field", validator=validators.is_file),
        ParameterInfo("flat_enabled", category="Flat-field", validator=bool),
        ParameterInfo("optimize_exposure_per", category="robust vs. speed"),
        ParameterInfo("rockit_distance", category="sample changer"),
        ParameterInfo(
            "optimize_attenuator",
            category="exposure/attenuator",
            validator=bool,
        ),
        ParameterInfo("optimize_mask_file", category="exposure/attenuator"),
        ParameterInfo("baguette_type", category="sample changer"),
        ParameterInfo("auto_stop_acc_mode", category="auto stop", validator=bool),
        ParameterInfo("auto_stop_workflow_path", category="auto stop"),
        ParameterInfo("auto_stop_threshold", category="auto stop", validator=float),
        ParameterInfo("auto_stop_scan_npoints", category="auto stop", validator=int),
        ParameterInfo("detector_saturation", category="auto stop", validator=float),
        ParameterInfo("auto_stop_target_max", category="auto stop", validator=float),
        ParameterInfo("auto_stop_target_min", category="auto stop", validator=float),
    ],
):
    # baguette types: step_size in motor user units (1 user unit = 13 mm)
    BAGUETTE_TYPE = {
        "standard": {
            "step_size": 1,  # 13 mm sample_holder pitch
            "number_of_samples": 16,
        },
        "wide": {
            "step_size": 24.8 / 13,  # 24.8 mm sample_holder pitch
            "number_of_samples": 9,
        },
    }

    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ):
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        defaults.setdefault("workflow", "streamline_without_calib_with_flat.json")
        defaults.setdefault("detector_name", "p3")
        defaults.setdefault("attenuator_name", "atten")
        defaults.setdefault("sample_changer_name", "streamline_sc")
        defaults.setdefault(
            "integration_options",
            {
                "method": "no_csr_ocl_gpu",
                "integrator_name": "sigma_clip_ng",
                "extra_options": {"max_iter": 3, "thres": 0},
                "error_model": "azimuthal",  # hybrid gives weird results
                "nbpt_rad": 4096,
                "unit": "q_nm^-1",
            },
        )
        defaults.setdefault("flat_enabled", True)
        defaults.setdefault("newflat", "/data/id31/inhouse/P3/flats.mat")
        defaults.setdefault("oldflat", "/data/id31/inhouse/P3/flats_old.mat")
        defaults.setdefault("optimize_pixel_value", 1e5)
        defaults.setdefault("optimize_nb_frames", 3)
        defaults.setdefault("optimize_min_exposure_time", 2)
        defaults.setdefault("optimize_max_exposure_time", 4)
        defaults.setdefault("optimize_exposure_per", "baguette")
        defaults.setdefault("rockit_distance", 0.07)
        defaults.setdefault("optimize_attenuator", True)
        defaults.setdefault("optimize_mask_file", None)
        defaults.setdefault("baguette_type", None)  # default: streamline_sc .yml config
        defaults.setdefault("auto_stop_acc_mode", False)
        defaults.setdefault("auto_stop_workflow_path", None)
        defaults.setdefault("auto_stop_threshold", 0.01)
        defaults.setdefault("auto_stop_scan_npoints", 10000)
        defaults.setdefault("detector_saturation", float(1 << 20))
        defaults.setdefault("auto_stop_target_max", None)
        defaults.setdefault("auto_stop_target_min", None)

        super().__init__(config=config, defaults=defaults)

        self._exposure_conditions: List[
            Union[optimize_exposure.ExposureCondition, None]
        ] = list()
        self._fixed_attenuator_position = None
        self._optimize_mask_array: Optional[numpy.ndarray] = None
        self._update_optimize_mask(self.optimize_mask_file)

        # upon Bliss restart push baguette type from Redis to streamline_scanner
        # object, which otherwise loads yml config (same as `baguette_type = None`)
        _btype = self.baguette_type
        if _btype is not None:
            self.baguette_type = _btype

    @property
    def baguette_type(self) -> Optional[str]:
        """Currently active baguette type name, returns None if unset."""
        return self._get_parameter("baguette_type")

    @baguette_type.setter
    def baguette_type(self, value: str) -> None:
        """Set the sample changer to use a different baguette type.

        Available types :
            - ``"standard"``: 16 samples, 13 mm pitch (step_size=1)
            - ``"wide"``: 9 samples, 24.8 mm pitch (step_size=24.8/13)

        :param baguette_type: baguette type name (see ``BAGUETTE_TYPE``)
        """
        if value not in self.BAGUETTE_TYPE:
            available = list(self.BAGUETTE_TYPE.keys())
            raise ValueError(f"Unknown baguette type '{value}'. Available: {available}")
        profile = self.BAGUETTE_TYPE[value]
        self.sample_changer.configure_baguette(**profile)
        self._set_parameter("baguette_type", value)
        print(
            f"Baguette type set to '{value}': "
            f"{profile['number_of_samples']} samples, "
            f"step_size={profile['step_size']:.4f} user units"
        )

    @property
    def optimize_exposure_per(self) -> Optional[str]:
        return self._get_parameter("optimize_exposure_per")

    @optimize_exposure_per.setter
    def optimize_exposure_per(self, value: Optional[str]):
        if value not in (None, "sample", "baguette"):
            raise ValueError("Allowed values are 'sample', 'baguette' or None")
        self._set_parameter("optimize_exposure_per", value)

    @property
    def optimize_mask_file(self) -> Optional[str]:
        return self._get_parameter("optimize_mask_file")

    @optimize_mask_file.setter
    def optimize_mask_file(self, filename: Optional[str]):
        self._update_optimize_mask(filename)

    def _update_optimize_mask(self, filename: Optional[str]):
        if not filename:
            self._optimize_mask_array = None
            self._set_parameter("optimize_mask_file", None)
            return

        try:
            self._optimize_mask_array = fabio.open(filename).data
        except Exception:
            print(f"Error: cannot load optimize_mask_file {filename}, reseting it!")
            self._optimize_mask_array = None
            self._set_parameter("optimize_mask_file", None)
            return

        self._set_parameter("optimize_mask_file", filename)

    @contextmanager
    def run_context(self):
        setup_globals.shopen(
            check_pilatus=False
        )  # check_pilatus = False when the detector was just started
        with super().run_context():
            yield

    def load(self):
        super().load()
        if self.optimize_exposure_per == "baguette" and not self.auto_stop_acc_mode:
            self.determine_exposure_conditions()

    def measure_sample(
        self, count_time: float = 1, *args, has_qrcode: bool = True, **kwargs
    ):
        if self.auto_stop_acc_mode:
            return self._measure_sample_auto_stop(count_time, *args, **kwargs)
        else:
            return self._measure_sample_default(
                count_time, *args, has_qrcode=has_qrcode, **kwargs
            )

    def _measure_sample_default(
        self, count_time: float = 1, *args, has_qrcode: bool = True, **kwargs
    ):
        with rockit(self.sample_changer.translation, self.rockit_distance):
            with self._optimize_sample_exposure(
                count_time, has_qrcode=has_qrcode
            ) as expo_time:
                expo_time_max = self.optimize_max_exposure_time
                expo_time_min = self.optimize_min_exposure_time
                expo_time = max(expo_time_min, min(expo_time, expo_time_max))

                if not self.dryrun:
                    ensure_shutter_open()
                try:
                    return super().measure_sample(expo_time, *args, **kwargs)
                except RuntimeError as e:
                    if "Pilatus protection" in str(e):
                        print(
                            f"Skip because of measurement error: {e}. Open shutter again"
                        )
                        return
                    raise

    def determine_exposure_conditions(self):
        """Pre-define optimal conditions: ascan at fixed attenuator position"""
        detector = getattr(setup_globals, self.detector_name)
        if self.default_attenuator is None:
            attenuator = getattr(setup_globals, self.attenuator_name)
            self.default_attenuator = attenuator.bits
        else:
            setup_globals.att(self.default_attenuator)
        self._exposure_conditions = optimize_exposure.optimal_exposure_conditions(
            *self.sample_changer.ascan_arguments(),
            detector,
            tframe=0.2,
            desired_counts=self.optimize_pixel_value,
            nframes_measure=1,
            nframes_default=self.optimize_nb_frames,
            reduce_desired_deviation=True,
            expose_with_integral_frames=False,
            mask=self._optimize_mask_array,
        )

    def determine_exposure_conditions_individually(self):
        """Pre-define optimal conditions: ct on each sample with adapted attenuator
        if the default attenuator position gives too much or too little counts"""
        detector = getattr(setup_globals, self.detector_name)
        attenuator = getattr(setup_globals, self.attenuator_name)
        att_value = attenuator.bits
        exposure_conditions = list()
        try:
            for _ in self.sample_changer.iterate_samples_without_qr():
                exposure_conditions.append(self._optimize_exposure_condition(detector))
        finally:
            setup_globals.att(att_value)
        self._exposure_conditions = exposure_conditions

    @contextmanager
    def _optimize_sample_exposure(
        self, count_time: float, has_qrcode: bool = True
    ) -> Generator[float, None, None]:
        """Selecting the optimal measurement conditions and returning the corresponding
        exposure time for the current sample."""

        if not self.optimize_exposure_per:
            # Optimization is disabled
            yield count_time
            return

        if self.optimize_exposure_per == "baguette":
            # Select pre-defined optimization if available
            count_time = self._set_exposure_condition()
            if count_time is not None:
                yield count_time
                return

        if not has_qrcode:
            # No QR-code probably means no sample so do not waste time optimizing
            yield count_time

        else:
            # Optimize condition for this sample individually
            detector = getattr(setup_globals, self.detector_name)
            attenuator = getattr(setup_globals, self.attenuator_name)
            att_value = attenuator.bits
            try:
                condition = self._optimize_exposure_condition(detector)
                yield condition.expo_time
            finally:
                setup_globals.att(att_value)

    def _set_exposure_condition(self) -> float | None:
        if not self._exposure_conditions:
            self.determine_exposure_conditions()
        sample_index = self.sample_changer.current_sample_index
        condition = self._exposure_conditions[sample_index]
        print(f"Pre-defined optimal exposure conditions: {condition}")
        if condition is None:
            return None
        setup_globals.att(condition.att_position)
        return condition.expo_time

    def _optimize_exposure_condition(
        self, detector
    ) -> optimize_exposure.ExposureCondition:
        return optimize_exposure.optimize_exposure_condition(
            detector,
            tframe=0.2,
            default_att_position=self.default_attenuator,
            desired_counts=self.optimize_pixel_value,
            dynamic_range=1 << 20,
            min_counts_per_frame=0,  # take 100
            nframes_measure=1,
            nframes_default=self.optimize_nb_frames,
            reduce_desired_deviation=True,
            expose_with_integral_frames=False,
            optimize_attenuator=self.optimize_attenuator,
            mask=self._optimize_mask_array,
        )

    def _measure_sample_auto_stop(self, count_time, *args, **kwargs):
        """Measure a sample using a loopscan + StopIntegrateSum preset.

        The loopscan runs until the supplied threshold is crossed or npoints is
        reached. Sums all accumulated frames and writes a ``*_sum.h5`` file.
        """
        if not self.auto_stop_workflow_path:
            raise RuntimeError("auto_stop_workflow_path is not set.")
        if not self.pyfai_config:
            raise RuntimeError("pyfai_config is not set.")
        if self.dryrun:
            print("Dry-run: skip auto-stop measurement")
            return None

        preset = StopIntegrateSum(
            workflow_threshold=self.auto_stop_threshold,
            workflow_path=self.auto_stop_workflow_path,
            detector_name=self.detector_name,
            detector_saturation=self.detector_saturation,
            pyfai_config_path=self.pyfai_config,
            attenuator_name=self.attenuator_name,
            frame_target_max=self.auto_stop_target_max,
            frame_target_min=self.auto_stop_target_min,
            ewoksjob_queue=self.queue,
        )
        with rockit(self.sample_changer.translation, self.rockit_distance):
            ensure_shutter_open()
            scan = setup_globals.loopscan(
                self.auto_stop_scan_npoints, count_time, *args, run=False, **kwargs
            )
            scan.acq_chain.add_preset(preset)
            scan.run()
        return scan

    def init_workflow(self, with_autocalibration: bool = False) -> None:
        if with_autocalibration:
            self.workflow = "streamline_with_calib_with_flat.json"
        else:
            self.workflow = "streamline_without_calib_with_flat.json"

        print(f"Active data processing workflow: {self.workflow}")

    def _job_arguments(self, scan_info, processed_metadata: dict):
        args, kwargs = super()._job_arguments(scan_info, processed_metadata)
        detector_name = self.detector_name

        kwargs["inputs"] += task_inputs(
            task_identifier="FlatFieldFromEnergy",
            inputs={
                "newflat": self.newflat,
                "oldflat": self.oldflat,
                "energy": getattr(setup_globals, self.energy_name).position,
                "enabled": self.flat_enabled and detector_name == "p3",
            },
        )

        # Rely on StreamlineScanner to set url and bliss_scan_url as for other SaveNexusPattern1D tasks
        kwargs["inputs"] += task_inputs(
            task_identifier="SaveNexusPattern1D",
            label="save_q_no_sigmaclip_hdf5",
            inputs={
                "nxprocess_name": f"{detector_name}_integrate_q_no_sigmaclip",
                "nxmeasurement_name": f"{detector_name}_integrated_q_no_sigmaclip",
                "metadata": {
                    f"{detector_name}_integrate_q_no_sigmaclip": {
                        "configuration": {"workflow": self.workflow}
                    }
                },
            },
        )

        # Use workflows from ewoksid31 module
        kwargs["load_options"] = {"root_module": "ewoksid31.workflows"}

        return args, kwargs


@contextmanager
def rockit(motor, distance):
    if distance:
        try:
            with setup_globals.rockit(motor, distance):
                print("ROCKING ON")
                yield
        finally:
            print("ROCKING OFF")
    else:
        print("ROCKING DISABLED")
        yield



import logging
from collections import OrderedDict
from collections.abc import MutableMapping
from copy import deepcopy
from typing import Any
from typing import Dict
from typing import Mapping

logger = logging.getLogger(__name__)

# Global in-memory Redis-like store
_REDIS_STORE: Dict[str, dict] = {}


class Connection:
    """Minimal stand-in for redis connection."""

    def exists(self, key: str) -> bool:
        return key in _REDIS_STORE


class LocalHashObjSetting(MutableMapping):
    """In-memory drop-in replacement for HashObjSetting."""

    connection = Connection()

    _LOCAL_REASON = None
    _LOCAL_WARNING_EMITTED = False

    def __init__(self, key: str):
        self.key = key

        if (
            not LocalHashObjSetting._LOCAL_WARNING_EMITTED
            and LocalHashObjSetting._LOCAL_REASON
        ):
            logger.warning(
                "Persistency is disabled and local in-memory storage is being used (%s)",
                LocalHashObjSetting._LOCAL_REASON,
            )
            LocalHashObjSetting._LOCAL_WARNING_EMITTED = True

    @property
    def _store(self) -> dict:
        return _REDIS_STORE.setdefault(self.key, {})

    def __repr__(self):
        return repr(self._store)

    def __getitem__(self, k: str) -> Any:
        return self._store[k]

    def __setitem__(self, k: str, value: Any) -> None:
        if value is None:
            self._store.pop(k, None)
        else:
            self._store[k] = deepcopy(value) if isinstance(value, Mapping) else value

    def __delitem__(self, k: str) -> None:
        del self._store[k]

    def __iter__(self):
        return iter(self._store)

    def __len__(self):
        return len(self._store)

    def get_all(self) -> dict:
        return deepcopy(self._store)

    def clear(self) -> None:
        _REDIS_STORE.pop(self.key, None)


class LocalOrderedHashObjSetting(LocalHashObjSetting):
    """
    Same as LocalHashObjSetting, but stores values in an OrderedDict so that
    insertion order is preserved.

    Behaves like a Redis hash that returns fields in insertion order.
    """

    @property
    def _store(self) -> OrderedDict:
        # Upgrade to OrderedDict if necessary
        store = _REDIS_STORE.get(self.key)

        if store is None:
            store = OrderedDict()
            _REDIS_STORE[self.key] = store
        else:
            if not isinstance(store, OrderedDict):
                _REDIS_STORE[self.key] = OrderedDict(store)
                store = _REDIS_STORE[self.key]

        return store



import numpy

from ...bliss_globals import (
    setup_globals,  # pyright: ignore[reportAttributeAccessIssue]
)
from ...version_utils import has_minimal_version
from .. import testing
from ..processors.stop_scan_xrpd_id31 import DemoStopIntegrateSum
from ..processors.stop_scan_xrpd_id31 import id31_patching


def _measure_baseline_imax(expo: float = 0.2) -> float:
    """Return the max pixel value from a single ct with no attenuation."""
    # id31_patching:
    # - ensures setup_globals.atten is available;
    # - ensures high energy so that attenuation with SiO2 performs as expected
    #   this is the id31.attenuator.Attenuator (`setup_globals.atten` on the beamline)
    # *note* it is called again in DemoStopIntegrateSum
    id31_patching(energy=75.05)

    setup_globals.atten.bits = 0
    scan = setup_globals.ct(expo, setup_globals.difflab6)
    return float(numpy.nanmax(scan.streams["difflab6:image"][:]))


@testing.integration_test
def test_attenuation_increase():
    """Verify that the preset increases attenuation when frames are saturating.

    Strategy: set ``detector_saturation`` below the measured baseline intensity
    so every frame appears saturating to the preset.  The workflow threshold is
    set astronomically high so the Ewoks workflow never fires a stop event and
    the scan runs to normal completion.  After ``npoints`` frames the preset
    must have moved ``atten.bits`` above 0.
    """
    if not has_minimal_version("bliss", "2.2"):
        testing.skip_integration_test("Requires bliss>=2.2")

    detector = setup_globals.difflab6
    expo = 0.2
    npoints = 5

    baseline_imax = _measure_baseline_imax(expo)

    # saturating threshold: set saturation well below the real detector output
    saturation = baseline_imax * 0.5
    p = DemoStopIntegrateSum(
        workflow_threshold=1e15,  # never trigger stop
        detector_name=detector.name,
        detector_saturation=saturation,
        frame_target_max=saturation * 0.9,
        frame_target_min=saturation * 0.1,
    )
    setup_globals.atten.bits = 0  # start with no attenuation

    scan = setup_globals.loopscan(npoints, expo, detector, run=False, save=False)
    scan.acq_chain.add_preset(p)
    scan.run()

    assert (
        setup_globals.atten.bits > 0
    ), f"Expected atten.bits > 0 after saturating frames, got {setup_globals.atten.bits}"


@testing.integration_test
def test_attenuation_decrease():
    """Verify that the preset decreases attenuation when frames are too dim.

    Strategy: start with high attenuation (``bits = 10``) and set
    ``frame_target_min`` well above the measured baseline intensity so every
    frame appears too dim.  ``detector_saturation`` is set high enough that
    removing attenuation is predicted to be safe.  The preset must lower
    ``atten.bits`` below the initial value.
    """
    if not has_minimal_version("bliss", "2.2"):
        testing.skip_integration_test("Requires bliss>=2.2")

    detector = setup_globals.difflab6
    expo = 0.2
    npoints = 5
    initial_bits = 10

    baseline_imax = _measure_baseline_imax(expo)

    # dim threshold: target_min above what the detector actually returns
    saturation = baseline_imax * 20
    p = DemoStopIntegrateSum(
        workflow_threshold=1e15,  # never trigger stop
        detector_name=detector.name,
        detector_saturation=saturation,
        frame_target_max=saturation * 0.9,
        frame_target_min=baseline_imax * 2,  # above measured baseline: always too dim
    )
    setup_globals.atten.bits = initial_bits

    scan = setup_globals.loopscan(npoints, expo, detector, run=False, save=False)
    scan.acq_chain.add_preset(p)
    scan.run()

    assert setup_globals.atten.bits < initial_bits, (
        f"Expected atten.bits < {initial_bits} after dim frames, "
        f"got {setup_globals.atten.bits}"
    )


@testing.integration_test
def stop_scan_pyfai_demo():
    if not has_minimal_version("bliss", "2.2"):
        testing.skip_integration_test("Requires bliss>=2.2")

    # set detector object and exposure time
    detector = setup_globals.difflab6
    expo = 0.2

    # create the preset; the Demo version sets up:
    # - the ID31 mock attenuator,
    # - demo PyFAI config,
    # - demo Ewoks workflow
    p = DemoStopIntegrateSum(workflow_threshold=0.005, detector_name=detector.name)

    # run a long loopscan with the preset - should stop at 10-15 frames
    scan = setup_globals.loopscan(100, expo, detector, run=False)
    scan.acq_chain.add_preset(p)
    scan.run()

    # check that the scan stopped early as expected
    npoints = len(scan.streams["difflab6:image"])
    assert npoints < 25, f"Stream 'difflab6:image' has {npoints} points, expected < 25"



from __future__ import annotations

import logging
import shutil
import sys
import tempfile
import time
import warnings
from collections.abc import Generator
from collections.abc import Sequence
from pathlib import Path

import redis
from blissdata.beacon.data import BeaconData
from ewoksjob.client import submit
from ewoksutils.task_utils import task_inputs

from blissoda.utils.directories import get_workflows_dir

from ..import_utils import unavailable_class
from ..import_utils import unavailable_module
from ..version_utils import assert_has_minimal_version

try:
    import gevent
except ImportError as ex:
    gevent = unavailable_module(ex)

try:
    from bliss import setup_globals  # type: ignore
except ImportError as ex:
    setup_globals = unavailable_class(ex)

try:
    from bliss.scanning.chain import AcquisitionChain  # type: ignore
except ImportError as ex:
    AcquisitionChain = unavailable_class(ex)

try:
    from bliss.scanning.chain import ChainIterationPreset  # type: ignore
except ImportError as ex:
    ChainIterationPreset = unavailable_class(ex)

try:
    from bliss.scanning.chain import ChainPreset  # type: ignore
except ImportError as ex:
    ChainPreset = unavailable_class(ex)


logger = logging.getLogger(__name__)

_MOTOR_POLL_INTERVAL = 0.2  # s to wait for filter/atten motor move
_CURSOR_READ_TIMEOUT = 0.5  # s to wait for a new blissdata frame
_REDIS_XREAD_BLOCK_MS = 100  # ms to block waiting for stream events
_ID31_ATTENUATOR_BITS = 32
_COLORS = {
    "red": "\033[31m",
    "green": "\033[32m",
    "yellow": "\033[33m",
    "blue": "\033[34m",
    "reset": "\033[0m",
}


def _color(text: str, color: str) -> str:
    """Wrap text in ANSI color codes only when stderr is a TTY."""
    if not hasattr(sys.stderr, "isatty") or not sys.stderr.isatty():
        return text
    return f"{_COLORS.get(color, '')}{text}{_COLORS['reset']}"


def _timestamp() -> str:
    now = time.time()
    ms = int((now - int(now)) * 1000)
    s = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(now)) + ",%03d" % ms
    return s


def _log_atten_action(
    level: int,
    symbol: str,
    color: str,
    frame_nb: int,
    imax: float,
    threshold: float,
    comparator: str,
    atten_msg: str,
) -> None:
    """Emit a log message for insertion/removal of filters or attenuators."""
    logger.log(
        level,
        "[%s] %s #%2.0f max: %2.2e %s %2.2e, %s",
        _timestamp(),
        _color(symbol, color),
        frame_nb,
        imax,
        comparator,
        threshold,
        _color(atten_msg, color),
    )


class StopIntegrateSum(ChainPreset):  # pyright: ignore[reportGeneralTypeIssues]
    def __init__(
        self,
        workflow_threshold: float,
        workflow_path: str | Path,
        detector_name: str,
        detector_saturation: float,
        pyfai_config_path: str | Path,
        attenuator_name: str = "filt_wheel",
        frame_target_max: float | None = None,
        frame_target_min: float | None = None,
        ewoksjob_queue: str = "mt",
    ) -> None:
        """An acquisition chain preset for automatic attenuation and scan stopping in
        XRPD measurements.

        To be used with:

        - a continous scan mimicking detector accumulation, e.g
          ``timescan(0.2, npoints=0)``.
        - an Ewoks workflow containing an ``IntegrateAndSendMsg`` task inheriting from
          ``ewoksxrpd.tasks.integrate._BaseIntegrateSingleGeometry``, and sending
          a {'type':'stop'} event via the `blissoda:feedback:stop_task` stream upon
          reaching some chosen `workflow_threshold`.

        Attach this preset to a scan to:

        - insert / remove a filter automatically based on specified min and max
          frame counts;
        - stop the scan based on the output of an Ewoks workflow triggered at scan
          start, collecting scan statistics.

        :param workflow_threshold: The "statistic" computed by the Ewoks workflow, upon
            reaching which the workflow sends a signal to stop the scan.
        :param workflow_path: Path to the Ewoks workflow `.json`.
        :param detector_name: Name of the detector used in the scan, whose `.image`
            counter (i.e. raw frames) will be fed to the Ewoks workflow.
        :param detector_saturation: Max of detector dynamic range. Damage above.
        :param pyfai_config_path: Path to the PyFAI configuration `.json`.
        :param attenuator_name: Name of the filter or attenuator Bliss object to
            insert/remove dynamically.
        :param frame_target_max: Maximum number of desired counts in a frame.
            Select attenuation to get closest to this value, and above
            `frame_target_min`. If None, set to `0.9 * detector_saturation`.
        :param frame_target_min: Minimum number of desired counts in a frame.
            Select attenuation to be above this value, and closest to
            `frame_target_max`. If None, set to `0.1 * detector_saturation`.
        :param ewoksjob_queue: Name of the Celery queue to route the Ewoks tasks
            submitted by the preset to. Default: 'mt'.
        """
        assert_has_minimal_version("bliss", "2.2")

        self._detector_saturation = detector_saturation
        self._workflow_threshold = workflow_threshold
        self._target_max = (
            frame_target_max
            if frame_target_max is not None
            else detector_saturation * 0.9
        )
        self._target_min = (
            frame_target_min
            if frame_target_min is not None
            else detector_saturation * 0.1
        )

        try:
            self._filter = getattr(setup_globals, attenuator_name)
        except AttributeError as ex:
            raise RuntimeError(
                f"The attenuator {attenuator_name} is not available "
                f"in setup_globals: {ex}."
            )
        self._filter_name = attenuator_name

        self._filename = None
        self._scan_number = None
        self._scan_key = None
        self._detector_name = detector_name

        self.pyfai_config = pyfai_config_path
        if self.pyfai_config is None:
            raise NotImplementedError(
                "pyfai_config cannot be None: workflow needs detector geometry params."
                "Could be supplied via _integration_options. TBD."
            )
        self._energy = None
        self._workflow_path = workflow_path
        self._integration_options = dict(
            method=("no", "csr", "cython"),
            unit="2th_deg",
            integrator_name="sigma_clip_ng",
            extra_options={"max_iter": 5, "thres": 3},
            error_model="azimuthal",
            correct_solid_angle=True,
            polarization_factor=0.97,
        )

        self._redis_url = BeaconData().get_redis_data_db()
        self._redis = redis.Redis.from_url(self._redis_url)
        self._stream_name = "blissoda:feedback:stop_task"
        self._last_id = "0-0"

        self._future = None
        self._events_buffer = []
        self._greenlet = None
        self._cursor = None
        self._queue = ewoksjob_queue

        self._transmissions_cache: tuple[tuple, Sequence] | tuple[()] = ()

    def __repr__(self) -> str:
        return (
            "StopSumPreset\n"
            f"  Detector:     {self._detector_name}\n"
            f"  Attenuator:   {self._filter_name}\n"
            f"  PyFAI config: {self.pyfai_config}\n"
            f"  Integration options: (precedence over config)\n"
            + "\n".join(
                [f"     {k}: {v}" for k, v in self._integration_options.items()]
            )
        )

    @property
    def pyfai_integration_options(self) -> dict:
        return self._integration_options

    @pyfai_integration_options.setter
    def pyfai_integration_options(self, value: dict) -> None:
        self._integration_options = value

    @property
    def pyfai_config(self) -> str | Path | None:
        return self._pyfai_config

    @pyfai_config.setter
    def pyfai_config(self, value: str | Path) -> None:
        if value is None:
            self._pyfai_config = None
            return
        p = Path(value)
        if not p.exists():
            raise FileNotFoundError(f"pyFAI config file does not exist: {p}")
        self._pyfai_config = str(p.absolute())

    def prepare(self, chain: AcquisitionChain) -> None:  # type: ignore
        self._cache_scan_info(chain)

    def start(self, chain: AcquisitionChain) -> None:  # type: ignore
        self._start_workflow()

    def stop(self, chain: AcquisitionChain) -> None:  # type: ignore
        self._stop_workflow()

    def _cache_scan_info(self, chain: AcquisitionChain) -> None:  # type: ignore
        """Called at scan.prepare()."""
        self._scan = chain.scan
        self._filename = chain.scan.scan_info["filename"]
        self._scan_number = chain.scan.scan_info["scan_nb"]
        self._scan_key = chain.scan._scan_data.key
        self._energy = chain.scan.scan_info["positioners"]["positioners_start"][
            "energy"
        ]

    def _get_workflow_task_inputs(self) -> list[dict]:
        if not self._pyfai_config:
            warnings.warn("No pyFAI config file specified")

        # pyfai config task inputs
        # out: energy / geometry (dict) / detector name(str)
        # detector config (dict) / mask (array|None) / ff (array|none)
        # integration options
        inputs = task_inputs(
            task_identifier="PyFaiConfig",
            inputs=dict(
                energy=self._energy,  # priority 1
                filename=self._pyfai_config,  # priority 4
                integration_options=self._integration_options,  # priority 2
            ),
        )

        # integration task inputs
        # uses iter_bliss_data
        inputs += task_inputs(
            task_identifier="IntegrateAndSendMsg",
            inputs=dict(
                scan_key=self._scan_key,  # checked first
                filename=self._filename,  # fall back to fname/scan
                scan_number=self._scan_number,
                detector_name=self._detector_name,
                response_stream_name=self._stream_name,
                acc_profile_threshold=self._workflow_threshold,
                redis_url=self._redis_url,
                retry_timeout=1000,  # 1s
            ),
        )

        # write frame sum task inputs
        # reads frames from output loopscan and writes a single summed frame
        # in *_sum.h5, then fed to the xrpd workflow
        inputs += task_inputs(
            task_identifier="WriteSumFrame",
            inputs=dict(
                filename=self._filename,
                scan_number=self._scan_number,
                detector_name=self._detector_name,
            ),
        )

        return inputs

    def _resolve_workflow_path(self) -> str:
        """Ensure the workflow file is in the proposal's workflows directory.

        Follows the same pattern as the blissoda xrpd processor: copy the
        source workflow into ``<processed>/workflows/`` (derived from the
        dataset filename) so that the Ewoks worker can always reach it via
        an absolute path on the shared filesystem.
        """
        src = Path(self._workflow_path).absolute()
        if not src.is_file():
            raise FileNotFoundError(f"Ewoks workflow not found: {src}")

        if self._filename:
            dst_dir = Path(get_workflows_dir(self._filename))
        else:
            dst_dir = Path(tempfile.gettempdir())
        dst = dst_dir / src.name

        if src != dst and not dst.is_file():
            dst_dir.mkdir(parents=True, exist_ok=True)
            shutil.copyfile(src, dst)

        return str(dst)

    def _start_workflow(self) -> None:
        """Called at scan.start(). Also starts _background_event_collector."""
        if not self._filename and not self._scan_key:
            raise RuntimeError("Neither scan key nor filename are known")

        # clear previous stream
        self._redis.delete(self._stream_name)
        self._last_id = "0-0"

        # submit workflow
        workflow = self._resolve_workflow_path()
        inputs = self._get_workflow_task_inputs()
        self._future = submit(
            args=(workflow,),
            kwargs=dict(
                inputs=inputs,
                convert_destination="last_stop_scan.json",
            ),
            queue=self._queue,
        )
        logger.info(
            "[%s] Workflow %s started for fname=%s key=%s, queue=%s",
            _timestamp(),
            self._future.uuid,
            self._filename,
            self._scan_key,
            self._queue,
        )

        # start background greenlet
        self._greenlet = gevent.spawn(self._background_event_collector)

    def _background_event_collector(self) -> None:
        """Continuously fetch and buffer feedback events."""
        while True:
            try:
                data = self._redis.xread(
                    {self._stream_name: self._last_id},
                    count=None,  # read all available
                    block=_REDIS_XREAD_BLOCK_MS,
                )

                if not data:
                    continue

                for _, events in data:
                    for event_id, event in events:
                        self._last_id = event_id
                        self._events_buffer.append(event)

            except Exception as ex:
                logger.exception("Error reading Redis stream: %s", ex)

    def _stop_workflow(self) -> None:
        """Called at scan.stop(). Only really stops _background_event_collector.
        The workflow needs to wait for scan.stop() to return in order to sum frames.
        """
        if self._greenlet is not None:
            self._greenlet.kill(block=True)
            self._greenlet = None

        # keep check only to catch exceptions: the WriteSumFrame task will
        # not fire until the data is flushed, which only happens at scan.stop()
        if self._future is None:
            return
        if self._future.done():
            try:
                self._future.result()
            except Exception as ex:
                logger.error("Stop scan workflow failed: %s", ex)

    def _get_filter_transmissions(self) -> Sequence:
        """Return a list of calculated transmission values for every filter or
        attenuator position.  Results are cached per preset instance and cleared
        when the energy or filter object is changed.
        """
        import numpy
        from bliss.common.auto_filter.filterset import FilterSet  # type: ignore
        from id31.attenuator import Attenuator  # pyright: ignore[reportMissingImports]
        from id31.attenuator import SiO2trans  # pyright: ignore[reportMissingImports]

        cache_key = (id(self._filter), self._energy)
        if self._transmissions_cache and self._transmissions_cache[0] == cache_key:
            return self._transmissions_cache[1]

        if isinstance(self._filter, FilterSet):
            result = [
                f.get("transmission_calc", None) for f in self._filter.get_filters()
            ]
        elif isinstance(self._filter, Attenuator):
            result = SiO2trans(self._energy, numpy.arange(_ID31_ATTENUATOR_BITS))
        else:
            raise RuntimeError(f"Filter {self._filter} is not supported.")

        self._transmissions_cache = (cache_key, result)
        return result

    def _find_best_filter_position(
        self, frame_max_intensity: float, current_atten_pos: int
    ) -> int | None:
        """Find the filter/attenuator position whose predicted intensity is closest
        to `_target_max` while staying below `_detector_saturation`.

        Returns `None` when the calculation cannot be performed
        (e.g. missing transmission data).
        """
        try:
            # get current transmission
            transmissions = self._get_filter_transmissions()
            t_current = transmissions[current_atten_pos]
            if t_current is None or t_current <= 0:
                return None

            # get unattenuated count rate
            count_rate = frame_max_intensity / t_current

            # find which transmission gives the cts closest to _target_max
            best_pos = None
            best_deviation = float("inf")
            for idx, t in enumerate(transmissions):
                if t is None or t < 0:
                    continue
                predicted = count_rate * t
                # only accept positions that give max cts below _detector_saturation
                if predicted < self._detector_saturation:
                    deviation = abs(predicted - self._target_max)
                    if deviation < best_deviation:
                        best_deviation = deviation
                        best_pos = idx
            return best_pos
        except Exception as ex:
            logger.warning("Cannot compute optimal filter position: %s", ex)
            return None

    def _predict_intensity_at(
        self, current_intensity_max: float, current_atten: int, target_atten: int
    ) -> float:
        """Predict the max pixel intensity at attenuator `target_atten` given a
        measurement at attenuator `current_atten`.  Returns `inf` when the prediction
        cannot be made, treated as saturation (conservative).
        """
        try:
            transmissions = self._get_filter_transmissions()
            t_current = transmissions[current_atten]
            t_target = transmissions[target_atten]
            if t_current and t_current > 0 and t_target is not None:
                return current_intensity_max * t_target / t_current
        except Exception:
            pass
        return float("inf")

    def _apply_filter(self, target_atten: int) -> None:
        """Move the filter/atten with blocking."""
        from bliss.common.auto_filter.filterset import FilterSet  # type: ignore

        try:
            if isinstance(self._filter, FilterSet):
                self._filter.set_filter(target_atten)
                while self._filter.rotation_axis.is_moving:
                    gevent.sleep(_MOTOR_POLL_INTERVAL)
            else:  # it's Attenuator
                self._filter.bits = target_atten
                # TODO proper way to block atten.bits?
                while self._filter.bits != target_atten:
                    gevent.sleep(_MOTOR_POLL_INTERVAL)
            logger.debug(
                "[%s] %s moved to position %d",
                _timestamp(),
                self._filter_name,
                target_atten,
            )
        except ValueError as err:
            logger.error(
                "[%s] Could not set %s to %d: %s",
                _timestamp(),
                self._filter_name,
                target_atten,
                err,
            )

    def get_iterator(
        self,
        chain: AcquisitionChain,  # type: ignore
    ) -> Generator[ChainIterationPreset, None, None]:  # type: ignore
        """Consume feedback events each iteration."""

        iter_nb = 0
        while True:
            yield ScanPointIterator(self, chain, iter_nb)
            iter_nb += 1


class ScanPointIterator(ChainIterationPreset):  # type: ignore
    def __init__(
        self,
        parent: StopIntegrateSum,
        chain: AcquisitionChain,  # type: ignore
        iter_nb: int,
    ):
        self.iter = iter_nb
        self.parent = parent
        self.chain = chain

    # executed synchronously
    def prepare(self) -> None:
        """Adjust attenuator/filter and stop scan if "stop" message received, at the
        preparation of each scan point. Skip the fist point.
        """
        if self.iter == 0:
            return

        self._adjust_filter()
        self._check_workflow_events()

    def start(self) -> None:
        """Start blissdata stream cursor at the start of the *first* scan point."""
        if self.iter == 0:
            stream = self.parent._scan.streams[f"{self.parent._detector_name}:image"]
            self.parent._cursor = stream.cursor()

    def _adjust_filter(self) -> None:
        """Read the frame acquired in the previous iter and adjust the
        attenuator/filter position.

        If:
        1. **Saturating** (max >= ``detector_saturation``): increase attenuation.
           Use transmission ratios to jump to the optimal position;
           fall back to +1.
        2. **Too dim** (max < ``frame_target_min``): decrease attenuation.
           Predict intensity at the target to avoid saturating;
           stay put if it would.
        3. **Acceptable range**: do nothing.
        """
        if self.parent._cursor is None:
            raise RuntimeError(
                "No blissdata Cursor present, did the first scan point run?"
            )

        view = self.parent._cursor.read(timeout=_CURSOR_READ_TIMEOUT, last_only=True)
        last_frame = view.get_data()
        imax_last_frame = float(last_frame.max())

        from bliss.common.auto_filter.filterset import FilterSet  # type: ignore

        if isinstance(self.parent._filter, FilterSet):
            current_atten = self.parent._filter.get_filter()
            n_filters = len(self.parent._filter.get_filters())
        else:  # Attenuator
            current_atten = self.parent._filter.bits
            n_filters = _ID31_ATTENUATOR_BITS
        max_atten = n_filters - 1

        # last frame was saturating: *increase* atten
        if (
            imax_last_frame >= self.parent._detector_saturation
            and current_atten < max_atten
        ):
            best = self.parent._find_best_filter_position(
                imax_last_frame, current_atten
            )
            if best is not None and best > current_atten:
                target_atten = best
            else:
                target_atten = current_atten + 1  # fallback: +1
            target_atten = min(target_atten, max_atten)

            _log_atten_action(
                logging.WARNING,
                "↓",
                "red",
                self.iter - 1,
                imax_last_frame,
                self.parent._detector_saturation,
                "≥",
                "atten: %d → %d" % (current_atten, target_atten),
            )
            self.parent._apply_filter(target_atten)

        # last frame was too dim: *decrease* atten
        elif imax_last_frame < self.parent._target_min and current_atten > 0:
            best = self.parent._find_best_filter_position(
                imax_last_frame, current_atten
            )
            if best is not None and best < current_atten:
                target_atten = best
            else:
                target_atten = current_atten - 1  # fallback: step down
            target_atten = max(target_atten, 0)

            # if removing atten would fry detector: do nothing
            predicted = self.parent._predict_intensity_at(
                imax_last_frame, current_atten, target_atten
            )
            if predicted >= self.parent._detector_saturation:
                _log_atten_action(
                    logging.WARNING,
                    "⨉",
                    "yellow",
                    self.iter - 1,
                    imax_last_frame,
                    self.parent._target_min,
                    "<",
                    "atten: %d (keep)" % current_atten,
                )
            else:
                _log_atten_action(
                    logging.WARNING,
                    "↑",
                    "blue",
                    self.iter - 1,
                    imax_last_frame,
                    self.parent._target_min,
                    "<",
                    "atten: %d → %d" % (current_atten, target_atten),
                )
                self.parent._apply_filter(target_atten)

        else:
            _log_atten_action(
                logging.INFO,
                "✓",
                "green",
                self.iter - 1,
                imax_last_frame,
                self.parent._target_max,
                "<",
                "atten: %d (in range)" % current_atten,
            )

    def _check_workflow_events(self) -> None:
        """Drain buffered workflow events and stop the scan if requested."""
        events_to_process, self.parent._events_buffer = (
            self.parent._events_buffer,
            [],
        )
        for event in events_to_process:
            if event.get(b"type") == b"stop":
                self.chain.scan.stop(wait=False)
                break



import numpy


def energy_wavelength(x):
    """keV to m and vice versa"""
    return 12.398419843320026 * 1e-10 / x


DEFAULT_CALIB = {
    "difflab6": {
        "dist": 5e-2,  # 5 cm
        "poni1": 10e-2,  # 10 cm
        "poni2": 10e-2,  # 10 cm
        "rot1": numpy.radians(10),  # 10 deg
        "rot2": 0,  # 0 deg
        "rot3": 0,  # 0 deg
        "energy": 12,  # 12 keV
        "detector": "Pilatus1M",
    },
    # TODO: produce different data or fix PONI
    "diffcam": {
        "dist": 5e-2,  # 5 cm
        "poni1": 10e-2,  # 10 cm
        "poni2": 10e-2,  # 10 cm
        "rot1": numpy.radians(10),  # 10 deg
        "rot2": 0,  # 0 deg
        "rot3": 0,  # 0 deg
        "energy": 12,  # 12 keV
        "detector": "detector",
        "detector_config": {"pixel1": 75e-6, "pixel2": 75e-6},
    },
}



"""Automatic pyfai integration for every scan with saving and plotting"""

import os
import tempfile
from glob import glob
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from ..persistent.parameters import ParameterInfo
from ..processor import BlissScanType
from ..utils.pyfai import read_config
from ..xrpd.processor import XrpdProcessor


class Bm20XrpdProcessor(
    XrpdProcessor,
    parameters=[
        ParameterInfo("pyfai_config_directory", category="PyFai"),
        ParameterInfo("integration_options", category="PyFai"),
    ],
):
    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        defaults.setdefault("pyfai_config_directory", tempfile.gettempdir())

        super().__init__(config=config, defaults=defaults)

    def _info_categories(self) -> Dict[str, dict]:
        categories = super()._info_categories()
        if not self.lima_names:
            return categories
        categories["PyFai"]["integration_options"] = "... (see below)"
        lima_name = self.lima_names[0]
        categories["PyFai integration"] = {
            "1. JSON file": self._detector_config_filename(lima_name, ".json"),
            "2. PONI file": self._detector_config_filename(lima_name, ".poni"),
            "3. User": self.integration_options,
            "Merged": self.get_integration_options(scan=None, lima_name=lima_name),
        }
        return categories

    def get_config_filename(self, lima_name: str) -> Optional[str]:
        return None

    def get_integration_options(
        self, scan: BlissScanType, lima_name: str
    ) -> Optional[dict]:
        options = self._default_azint_options(lima_name)
        options.update(self._default_calib_options(lima_name))
        integration_options = self.integration_options
        if integration_options:
            options.update(integration_options.to_dict())
        return options

    def _default_azint_options(self, lima_name: str) -> dict:
        filename = self._detector_config_filename(lima_name, ".json")
        return read_config(filename)

    def _default_calib_options(self, lima_name: str) -> dict:
        filename = self._detector_config_filename(lima_name, ".poni")
        return read_config(filename)

    def _detector_config_filename(self, lima_name: str, ext: str) -> Optional[str]:
        pyfai_config_directory = self.pyfai_config_directory
        if not pyfai_config_directory:
            pyfai_config_directory = tempfile.gettempdir()
        pattern = os.path.join(pyfai_config_directory, lima_name, f"*{ext}")
        files = sorted(glob(pattern))
        if not files:
            return
        return files[-1]

    def get_inputs(self, scan, lima_name: str) -> List[dict]:
        inputs = super().get_inputs(scan, lima_name)
        return inputs



from __future__ import annotations

import json
import logging
from pathlib import Path
from typing import Any
from typing import Dict

from blissdata.beacon.data import BeaconData
from blissdata.redis_engine.store import DataStore

from blissoda.import_utils import unavailable_class
from blissoda.import_utils import unavailable_function

from ..persistent.parameters import ParameterInfo

try:
    from bliss.scanning.scan import ScanState as BlissScanState
    from tomo.globals import get_active_tomo_config
except ImportError as ex:
    BlissScanState = unavailable_class(ex)
    get_active_tomo_config = unavailable_function(ex)

from esrf_pathlib import ESRFPath
from ewoksjob.client import submit

from blissoda.resources import resource_filename
from blissoda.tomo.online_tomo_plotter import OnlineTomoAccumulatedPlotter
from blissoda.tomo.tomo_processor import TomoProcessor

from .tomo_model import TomoProcessorModel
from .utils import ImageKey
from .utils import calculate_CoR_estimate
from .utils import get_estimate_cor_metadata
from .utils import get_reconstruction_metadata

logger = logging.getLogger(__name__)

FUTURE_TIMEOUT = None
SUPPORTED_TOMO_SCANS = ["tomo:basic", "tomo:zseries", "tomo:fullturn", "tomo:halfturn"]
# The following scan types are not yet supported:
# tomo:helical, tomo:zhelical, tomo:ptychotomo, tomo:multitomo , tomo:holotomo


class OnlineTomoSliceProcessor(
    TomoProcessor,
    parameters=[
        ParameterInfo(
            "batch_size",
            category="slice_reconstruction_parameters",
            doc="Number of projections to process in each batch",
        ),
        ParameterInfo(
            "flat_reduction_method",
            category="flat_dark_reduction_parameters",
            doc="Method for flat field reduction ('mean', 'median')",
        ),
        ParameterInfo(
            "padding_mode",
            category="slice_reconstruction_parameters",
            doc="Padding mode for reconstruction ('edges', 'zeros', etc.)",
        ),
        ParameterInfo(
            "extra_options",
            category="slice_reconstruction_parameters",
            doc="Additional options for reconstruction as a dictionary",
        ),
    ],
):
    def __init__(
        self,
        config: Dict[str, Any] | None = None,
        defaults: Dict[str, Any] | None = None,
    ) -> None:
        """
        Initialize the OnlineTomoSliceProcessor for online tomographic slice reconstruction.

        This processor can be integrated into a BLISS beamline configuration. Typical
        usage involves adding the following to a configuration yaml file:

        - name: tomo_blissoda
          plugin: generic
          class: OnlineTomoSliceProcessor
          package: blissoda.tomo.online_tomo_slice_processor
        """

        if defaults is None:
            defaults = {}
        defaults.setdefault("trigger_at", "PREPARED")
        defaults.setdefault("workflow", "tomo_reconstruction_online.json")
        defaults.setdefault("nabu_config_file", None)
        defaults.setdefault("slice_index", "middle")
        defaults.setdefault("cor_algorithm", "sliding-window")
        defaults.setdefault("phase_retrieval_method", "None")
        defaults.setdefault("delta_beta", "100")
        defaults.setdefault("_flip_left_right", False)
        defaults.setdefault("offset_mm", 0.0)
        defaults.setdefault("_orientation_factor", 1.0)
        defaults.setdefault("estimated_cor", 0.0)
        defaults.setdefault("show_last_slice", False)
        defaults.setdefault("flat_reduction_method", "mean")
        defaults.setdefault("batch_size", 100)
        defaults.setdefault("padding_mode", "edges")
        defaults.setdefault("extra_options", None)

        super(TomoProcessor, self).__setattr__(
            "_tomo_model", TomoProcessorModel(**defaults)
        )

        # Initialize internal state
        self._sequence_scan = None
        self._dark_n = 0
        self._flat_n = 0
        self._sample_name = None

        # Initialize workflow output paths
        self.reduced_dark_path = None
        self.reduced_flat_path = None

        # Initialize workflow futures
        self.reduced_dark_future = None
        self.reduced_flat_future = None

        # Workflow paths
        self._reduced_dark_flat_workflow = "tomo_reduction_online.json"
        self._reconstruction_workflow = "tomo_reconstruction_online.json"

        # Initialize the plotter for accumulated slices
        self.accumulated_plotter = OnlineTomoAccumulatedPlotter(
            history=1,
        )

        super(TomoProcessor, self).__init__(config=config, defaults=defaults)

    def execute_workflow(self, scan) -> None:
        """Execute workflow based on scan type and state."""

        # Check if this is a tomography configuration scan
        technique = scan.scan_info.get("technique", "")
        if (
            "tomoconfig" in technique
            and scan.scan_info["title"] in SUPPORTED_TOMO_SCANS
        ):
            logger.info(
                f"Starting tomography sequence for scan: {scan.scan_info.get('title', 'Unknown')}"
            )
            self._reset_state()
            self._sequence_scan = scan
            self._output_path = self.get_processed_path(scan.scan_saving.data_path)
            # Set sample name from output path, format will be {sample}_{dataset}
            self._sample_name = self._output_path.name
            return

        # Handle individual scans within the sequence
        if self._sequence_scan is not None:
            scan_type = self._get_scan_type(scan)

            logger.info(f"Processing scan: {scan_type}")

            if scan_type is ImageKey.DARK_FIELD:
                self._dark_n = self._extract_frame_count(scan, "dark")
                self.on_new_darkflat_scan("dark", index=self._flat_n)
            elif scan_type is ImageKey.FLAT_FIELD:
                self._flat_n = self._extract_frame_count(scan, "flat")
                self.on_new_darkflat_scan("flat", index=self._dark_n)
            elif scan_type is ImageKey.PROJECTION:
                self.wait_for_futures()
                # Only run reconstruction if we have both dark and flat processed
                if self.reduced_dark_path and self.reduced_flat_path:
                    self.result = self.on_new_projection_scan()
                else:
                    logger.warning(
                        "Cannot run reconstruction: missing reduced dark or flat files"
                    )

    def _reset_state(self) -> None:
        """Reset internal state for new tomography sequence."""
        self._sequence_scan = None
        self._dark_n = 0
        self._flat_n = 0
        self._output_path = None
        self._sample_name = None
        self.reduced_dark_path = None
        self.reduced_flat_path = None
        self.reduced_dark_future = None
        self.reduced_flat_future = None

    def _extract_frame_count(self, scan, frame_type: str) -> int:
        technique = scan.scan_info.get("technique", {})
        frame_info = technique.get(frame_type, {})
        return int(frame_info.get(f"{frame_type}_n", 0))

    def _get_scan_type(self, scan) -> ImageKey:
        """Determine the type of scan based on the key in scan_info."""
        image_key = int(scan.scan_info.get("technique", {}).get("image_key", 3))
        return ImageKey(image_key)

    def estimate_CoR(self) -> float:
        """
        Estimate the center of rotation based on scan metadata and processor parameters.
        """

        pixel_size_mm, detector_width, translation_y_mm = get_estimate_cor_metadata()

        center_of_rotation = calculate_CoR_estimate(
            pixel_size_mm=pixel_size_mm,
            detector_width=detector_width,
            translation_y_mm=translation_y_mm,
            offset_mm=self.offset_mm,
            flip=self._flip_left_right,
            orientation_factor=self._orientation_factor,
        )
        return center_of_rotation

    def wait_for_futures(self):
        """Wait for any ongoing reduction workflows to complete."""
        for attr in ["reduced_dark_future", "reduced_flat_future"]:
            future = getattr(self, attr, None)
            if future is not None:
                try:
                    future.result(timeout=FUTURE_TIMEOUT)  # Wait for completion
                    logger.info(f"{attr} completed successfully")
                except Exception as e:
                    logger.error(f"Error in {attr}: {e}")

    def get_processed_path(self, data_path: str) -> Path:
        """
        Get the processed path for the reconstructed slices.
        """
        output_path = ESRFPath(data_path)
        return (
            output_path.processed_data_path / output_path.collection / output_path.stem
        )

    def _get_workflow(self, workflow: str) -> Dict[str, Any]:
        """Load workflow definition from JSON file."""
        try:
            with open(resource_filename("tomo", workflow), "r") as wf:
                return json.load(wf)
        except Exception as e:
            logger.error(f"Failed to load workflow {workflow}: {e}")
            raise

    def _connect_data_store(self) -> None:
        """Initialize connection to the data store."""
        try:
            beacon_client = BeaconData()
            redis_url = beacon_client.get_redis_data_db()
            data_store = DataStore(redis_url)
            logger.info("Connected to beacon")
            return data_store
        except Exception as e:
            logger.error(f"Failed to connect to beacon: {e}")
            raise

    def on_new_darkflat_scan(self, scan_type: str, index: int = 0):
        """Run reduction workflow for dark or flat frames."""
        try:
            data_store = self._connect_data_store()
            _, key = data_store.get_last_scan()

            # Create output path
            output_path = self._output_path / f"references/{scan_type}_field.h5"
            output_path.parent.mkdir(parents=True, exist_ok=True)

            # Determine reduction method based on scan type
            if "dark" in scan_type.lower():
                reduction_method = "mean"
                convert_destination = str(
                    output_path.parent
                    / f"workflows/{self._sample_name}_dark_online.json"
                )
            elif "flat" in scan_type.lower():
                reduction_method = self.flat_reduction_method
                convert_destination = str(
                    output_path.parent
                    / f"workflows/{self._sample_name}_flat_online.json"
                )
            else:
                logger.warning(f"Unknown scan type for reduction: {scan_type}")
                return

            logger.info(
                f"Submitting reduction workflow for {scan_type} with method {reduction_method}"
            )

            # Submit the reduction workflow
            future = self._submit_reduction_workflow(
                scan_key=key,
                output_path=output_path,
                reduction_method=reduction_method,
                index=index,
                convert_destination=convert_destination,
            )

            # Use dynamic attribute names to store futures and paths
            # e.g., reduced_dark_future, reduced_flat_future
            attr_name = f"reduced_{scan_type}_future"
            setattr(self, attr_name, future)
            # e.g., reduced_dark_path, reduced_flat_path
            attr_name = f"reduced_{scan_type}_path"
            setattr(self, attr_name, str(output_path))

        except Exception as e:
            logger.error(f"Failed to run reduction workflow for {scan_type}: {e}")

    def on_new_projection_scan(self):
        """Run reconstruction workflow for projection data."""
        try:
            data_store = self._connect_data_store()
            _, key = data_store.get_last_scan()
            scan = data_store.load_scan(key)

            # Create output path
            output_path = self._output_path / "slices/online"
            output_path.mkdir(parents=True, exist_ok=True)

            # Get rotation motor name from tomo configuration
            tomo_config = get_active_tomo_config()
            rotation_motor = tomo_config.rotation_axis.name

            # Extract reconstruction parameters from scan metadata
            technique_info = scan.info.get("technique", {})
            tomo_n = int(technique_info.get("proj", {}).get("proj_n"))
            scan_info = self._sequence_scan.scan_info.get("technique", {}).get(
                "scan", {}
            )
            halftomo = scan_info.get("half_acquisition", False)

            # Get metadata for CoR estimation
            center_of_rotation = self.estimate_CoR()

            # Get reconstruction metadata
            distance_m, energy_keV, pixel_size_m = get_reconstruction_metadata(
                self._sequence_scan
            )

            logger.info(
                f"Submitting reconstruction workflow - tomo_n: {tomo_n}, rotation_motor: {rotation_motor}"
            )

            # Submit the reconstruction workflow
            job = self._submit_reconstruction_workflow(
                scan_key=key,
                output_path=output_path,
                rotation_motor=rotation_motor,
                total_nb_projection=tomo_n,
                center_of_rotation=center_of_rotation,
                batch_size=self.batch_size,
                reduced_dark_path=self.reduced_dark_path,
                reduced_flat_path=self.reduced_flat_path,
                pixel_size_m=pixel_size_m,
                distance_m=distance_m,
                energy_keV=energy_keV,
                slice_index=self.slice_index,
                delta_beta=self.delta_beta,
                halftomo=halftomo,
                padding_mode=self.padding_mode,
                extra_options=self.extra_options,
            )

            logger.info("Reconstruction workflow submitted successfully")

            # Trigger the plotter if show_last_slice is enabled
            if self.show_last_slice:
                self.accumulated_plotter.handle_workflow_result(
                    future=job,
                    output_path=str(output_path),
                    slice_index=self.slice_index,
                    batch_size=self.batch_size,
                )

            return job

        except Exception as e:
            logger.error(f"Failed to run reconstruction workflow: {e}")

    def _submit_reduction_workflow(
        self,
        scan_key: str,
        output_path: Path,
        reduction_method: str,
        index: int,
        convert_destination: str,
    ):
        """Submit a reduction workflow for dark or flat frames."""

        # Load workflow definition
        workflow = self._get_workflow(self._reduced_dark_flat_workflow)

        # Prepare inputs for the workflow
        inputs = [
            {
                "task_identifier": "ewokstomo.tasks.online.reducedarkflat.OnlineReduceDarkFlat",
                "name": "scan_key",
                "value": scan_key,
            },
            {
                "task_identifier": "ewokstomo.tasks.online.reducedarkflat.OnlineReduceDarkFlat",
                "name": "output_file_path",
                "value": str(output_path),
            },
            {
                "task_identifier": "ewokstomo.tasks.online.reducedarkflat.OnlineReduceDarkFlat",
                "name": "reduction_method",
                "value": reduction_method,
            },
            {
                "task_identifier": "ewokstomo.tasks.online.reducedarkflat.OnlineReduceDarkFlat",
                "name": "index",
                "value": index,
            },
        ]

        # Submit the workflow
        kwargs = {
            "inputs": inputs,
            "convert_destination": convert_destination,
        }

        try:
            job = submit(args=(workflow,), kwargs=kwargs, queue=self.queue)
            logger.info(f"Submitted reduction workflow job: {job}")
            return job
        except Exception as e:
            logger.error(f"Failed to submit reduction workflow: {e}")
            raise

    def _submit_reconstruction_workflow(
        self,
        scan_key: str,
        output_path: Path,
        rotation_motor: str,
        batch_size: int,
        total_nb_projection: int,
        center_of_rotation: float,
        reduced_dark_path: str,
        reduced_flat_path: str,
        pixel_size_m: float,
        distance_m: float,
        energy_keV: float,
        slice_index: str = "middle",
        delta_beta: float = 100.0,
        halftomo: bool = False,
        padding_mode: str = "edges",
        extra_options: dict | None = None,
    ):
        """Submit a reconstruction workflow for projection data."""

        if extra_options is None:
            extra_options = {"centered_axis": True}

        # Load workflow definition
        workflow = self._get_workflow(self._reconstruction_workflow)

        convert_destination = str(
            output_path.parent
            / f"workflows/{self._sample_name}_slice_reconstruction.json"
        )

        # Prepare inputs for the workflow
        inputs = [
            {
                "task_identifier": "ewokstomo.tasks.online.reconstruct_slice.OnlineReconstructSlice",
                "name": "scan_key",
                "value": scan_key,
            },
            {
                "task_identifier": "ewokstomo.tasks.online.reconstruct_slice.OnlineReconstructSlice",
                "name": "output_path",
                "value": str(output_path),
            },
            {
                "task_identifier": "ewokstomo.tasks.online.reconstruct_slice.OnlineReconstructSlice",
                "name": "rotation_motor",
                "value": rotation_motor,
            },
            {
                "task_identifier": "ewokstomo.tasks.online.reconstruct_slice.OnlineReconstructSlice",
                "name": "total_nb_projection",
                "value": total_nb_projection,
            },
            {
                "task_identifier": "ewokstomo.tasks.online.reconstruct_slice.OnlineReconstructSlice",
                "name": "center_of_rotation",
                "value": center_of_rotation,
            },
            {
                "task_identifier": "ewokstomo.tasks.online.reconstruct_slice.OnlineReconstructSlice",
                "name": "batch_size",
                "value": batch_size,
            },
            {
                "task_identifier": "ewokstomo.tasks.online.reconstruct_slice.OnlineReconstructSlice",
                "name": "reduced_dark_path",
                "value": str(reduced_dark_path),
            },
            {
                "task_identifier": "ewokstomo.tasks.online.reconstruct_slice.OnlineReconstructSlice",
                "name": "reduced_flat_path",
                "value": str(reduced_flat_path),
            },
            {
                "task_identifier": "ewokstomo.tasks.online.reconstruct_slice.OnlineReconstructSlice",
                "name": "pixel_size_m",
                "value": pixel_size_m,
            },
            {
                "task_identifier": "ewokstomo.tasks.online.reconstruct_slice.OnlineReconstructSlice",
                "name": "distance_m",
                "value": distance_m,
            },
            {
                "task_identifier": "ewokstomo.tasks.online.reconstruct_slice.OnlineReconstructSlice",
                "name": "energy_keV",
                "value": energy_keV,
            },
            {
                "task_identifier": "ewokstomo.tasks.online.reconstruct_slice.OnlineReconstructSlice",
                "name": "slice_index",
                "value": slice_index,
            },
            {
                "task_identifier": "ewokstomo.tasks.online.reconstruct_slice.OnlineReconstructSlice",
                "name": "delta_beta",
                "value": delta_beta,
            },
            {
                "task_identifier": "ewokstomo.tasks.online.reconstruct_slice.OnlineReconstructSlice",
                "name": "halftomo",
                "value": halftomo,
            },
            {
                "task_identifier": "ewokstomo.tasks.online.reconstruct_slice.OnlineReconstructSlice",
                "name": "padding_mode",
                "value": padding_mode,
            },
            {
                "task_identifier": "ewokstomo.tasks.online.reconstruct_slice.OnlineReconstructSlice",
                "name": "extra_options",
                "value": extra_options,
            },
        ]

        # Submit the workflow
        kwargs = {
            "inputs": inputs,
            "convert_destination": convert_destination,
        }

        try:
            job = submit(args=(workflow,), kwargs=kwargs, queue=self.queue)
            logger.info(f"Submitted reconstruction workflow job: {job}")
            return job
        except Exception as e:
            logger.error(f"Failed to submit reconstruction workflow: {e}")
            raise



from typing import Any
from typing import Dict
from typing import Optional

from ..exafs import scan_utils
from ..exafs.processor import ExafsProcessor


class Bm08ExafsProcessor(ExafsProcessor):

    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
    ) -> None:
        if defaults is None:
            defaults = {}

        defaults.setdefault(
            "workflow", "/home/opd08/ewoks/resources/dau_bm08/online_exafs_plots.ows"
        )
        defaults.setdefault("queue", "online")
        defaults.setdefault("_scan_type", "Kscan")
        counters = defaults.setdefault("_counters", dict())
        counters.setdefault(
            "Kscan",
            {
                "mu_name": "I1_eh2",
                "mon_name": "I0_eh2",
                "min_log": True,
                "energy_name": "energy",
                "energy_unit": "keV",
            },
        )

        super().__init__(config=config, defaults=defaults)

    def _scan_type_from_scan(self, scan: scan_utils.ScanType) -> Optional[str]:
        scan_parameters = scan.scan_info.get("scan_parameters", {})
        return scan_parameters.get("scan_type")

    def _multi_xas_scan(self, scan: scan_utils.ScanType) -> bool:
        return False

    def _multi_xas_subscan_size(self, scan: scan_utils.ScanType) -> int:
        raise NotImplementedError



from typing import Any
from typing import Dict

from ..version_utils import has_minimal_version


def adapt_legacy_metadata(metadata: Dict[str, Any]) -> None:
    """Modifies ``metadata`` in case the installed ``icat-esrf-definitions``
    library does not support the latest ICAT metadata schema.
    """
    if not has_minimal_version("icat-esrf-definitions", "2.0.0"):
        _ = metadata.pop("technique_pid")



from __future__ import annotations

import logging
import time
from collections import OrderedDict
from pathlib import Path
from typing import Optional

import numpy
import silx

from blissoda.tomo.utils import apply_labels
from blissoda.tomo.utils import compute_axes

from ..bliss_globals import current_session
from ..flint import capture_errors
from ..flint.plotter import BasePlotter
from ..import_utils import unavailable_class

try:
    from flint.viewers.custom_image.client import ImageView
except ImportError as ex:
    ImageView = unavailable_class(ex)

_logger = logging.getLogger(__name__)


class SingleSliceImshow(BasePlotter):
    """
    Manage a Flint window showing the most recent reconstructed slice with enhanced controls,
    using the flint DataPlot–based ImageView.
    """

    TITLE = "Last Reconstructed Slice"

    def __init__(self, history: int = 1) -> None:
        super().__init__(max_plots=history)
        self._cache: OrderedDict[str, numpy.ndarray] = OrderedDict()

    def _load_and_show(self, img_path):
        with silx.io.open(img_path) as h5In:
            img = h5In["entry0000/reconstruction/results/data"][:]
        self.set_image(numpy.squeeze(img))

    def monitor_and_display_slice(self, output_path: str) -> None:
        """
        Start monitoring the slice output directory and display the first reconstructed
        slice as soon as an HDF5 file becomes available.
        """

        output_path = Path(output_path)
        output_dir = output_path.parent.parent / "slices"
        file_prefix = output_path.stem
        if len(list(output_dir.glob(f"{file_prefix}*_xy_*.h*5"))) > 0:
            _logger.warning("Slice file already exists in %s", output_dir)
        img_path = self._wait_for_slice_file(output_dir, file_prefix=file_prefix)
        if img_path is not None:
            self._load_and_show(img_path)

    def _wait_for_slice_file(
        self,
        output_dir: Path,
        file_prefix: str,
        poll_s: float = 0.5,
        timeout_s: float = 600.0,
    ) -> Optional[str]:
        """Wait for an HDF5 slice file to appear in the output directory."""
        deadline = time.monotonic() + timeout_s

        while time.monotonic() < deadline:
            files = list(output_dir.glob(f"{file_prefix}*_xy_*.hdf5"))
            time.sleep(poll_s)
            if files:
                return str(files[0])

        _logger.warning("No slice HDF5 found in %s after %.1fs", output_dir, timeout_s)
        return None

    @capture_errors
    def set_image(self, image: numpy.ndarray) -> None:
        """
        Display a new image with physical axis limits based on pixel size and image center.
        """
        widget = self._get_plot(self.TITLE, ImageView)
        self._set_title(widget)

        x_axis, y_axis = compute_axes(image)
        origin = (float(x_axis[0]), float(y_axis[0]))

        dx = float(x_axis[1] - x_axis[0]) if len(x_axis) > 1 else 1.0
        dy = float(y_axis[1] - y_axis[0]) if len(y_axis) > 1 else 1.0
        scale = (dx, dy)

        widget.set_data(image, origin=origin, scale=scale)
        apply_labels(widget)

        self._cache[self.TITLE] = image
        self.purge_tasks()
        self._purge()

    def _set_title(self, widget: ImageView) -> None:
        """
        Set the plot window title using the current data filename from scan saving.
        """
        widget.title = current_session.scan_saving.data_filename

    def _purge(self) -> None:
        """
        Remove oldest images beyond history.
        """
        while len(self._cache) > self._max_plots:
            self._cache.popitem(last=False)



from ...import_utils import unavailable_class
from ...version_utils import has_minimal_version

if has_minimal_version("blissdata", "1"):
    try:
        import bliss  # noqa F401
    except ImportError as ex:
        PersistentNdArray = unavailable_class(ex)
    else:
        from .ndarrayv1 import PersistentNdArray
else:
    from .ndarrayv0 import PersistentNdArray  # noqa F401



"""User API for HDF5 conversion on the Bliss repl"""

import os
import shutil
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from ewoksjob.client import submit

from ..bliss_globals import current_session
from ..persistent.parameters import ParameterInfo
from ..processor import BaseProcessor
from ..processor import BlissScanType
from ..resources import resource_filename
from ..utils import counters
from ..utils import directories


def _coefficients_validator(value):
    if value is None:
        return None
    if not isinstance(value, (list, tuple)):
        raise TypeError("Coefficients must be a list or tuple")
    if len(value) < 2:
        raise ValueError("Coefficients must be a list of 2 or more values")
    return list(value)


def _map_shape_validator(value):
    if value is None:
        return None
    if not isinstance(value, (list, tuple)):
        raise TypeError("Shape must be a list or tuple")
    if len(value) != 2:
        raise ValueError("Shape must have two numbers: (nslow, nfast)")
    return tuple(value)


def _positive_number_validator(value):
    if value is None:
        return None
    if value < 0:
        raise ValueError("Value cannot be negative")
    return value


def _path_validator(value):
    if value is None:
        return None
    if not os.path.exists(value):
        raise FileNotFoundError(value)
    return value


class Id16bXeolProcessor(
    BaseProcessor,
    parameters=[
        ParameterInfo("workflow", category="workflows"),
        ParameterInfo("queue", category="workflows", deprecated_names=["worker"]),
        ParameterInfo("main_counter", category="data"),
        ParameterInfo(
            "calibration_x",
            category="calibration",
            validator=_coefficients_validator,
            doc="Spectral energy calibration polynomial coefficients.",
        ),
        ParameterInfo(
            "calibration_y",
            category="calibration",
            validator=_coefficients_validator,
            doc="Spectral intensity calibration polynomial coefficients.",
        ),
        ParameterInfo(
            "threshold",
            category="filters",
            validator=_positive_number_validator,
            doc="Total spectral intensity threshold.",
        ),
        ParameterInfo("normalization_counter", category="normalization"),
        ParameterInfo(
            "configuration_path", validator=_path_validator, category="PyMCA"
        ),
        ParameterInfo("output_suffix", category="results"),
        ParameterInfo("data_portal_upload", category="data portal"),
    ],
):
    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
    ) -> None:
        if defaults is None:
            defaults = {}

        defaults.setdefault("trigger_at", "END")
        defaults.setdefault("queue", "celery")
        defaults.setdefault("calibration_x", None)  # default could be None of [0,1]
        defaults.setdefault("calibration_y", None)  # default could be None of [0,1]
        defaults.setdefault("threshold", 0)
        defaults.setdefault("workflow", resource_filename("id16b", "id16b_xeol.json"))
        defaults.setdefault("output_suffix", "_xeol")

        super().__init__(config=config, defaults=defaults)

    def on_new_scan_metadata(self, scan: BlissScanType) -> None:
        if not self.scan_requires_processing(scan):
            return
        kwargs = self.get_submit_arguments(scan)
        workflow = self.get_workflow(scan)

        with_saving = scan.scan_info.get("save")
        if with_saving:
            kwargs["convert_destination"] = self._get_workflow_destination(scan)

        upload_parameters = None
        if self.data_portal_upload and with_saving:
            # It is allowed to upload the same processed directory more than once.
            upload_parameters = self._get_workflow_upload_parameters(scan)
        if upload_parameters:
            kwargs["upload_parameters"] = upload_parameters

        _ = submit(args=(workflow,), kwargs=kwargs, queue=self.queue)

    def _trigger_workflow_on_new_scan(self, scan: BlissScanType) -> None:
        self.on_new_scan_metadata(scan)

    def scan_requires_processing(self, scan: BlissScanType) -> bool:
        if not scan.scan_info.get("save"):
            return False
        if scan.scan_info["type"] != "fscan2d":
            return False
        if not self._has_xeol_controllers(scan):
            return False
        return True

    def _has_xeol_controllers(self, scan: BlissScanType) -> bool:
        controller_names = counters.counter_to_controller_names([self.main_counter])
        ctrl_prefixes = [f"{ctrl_name}:" for ctrl_name in controller_names]
        return any(
            name
            for name in scan.scan_info["channels"].keys()
            if any(name.startswith(ctrl_prefix) for ctrl_prefix in ctrl_prefixes)
        )

    def get_submit_arguments(self, scan: BlissScanType) -> dict:
        return {
            "inputs": self.get_inputs(scan),
            "outputs": [{"all": False}],
        }

    def get_filename(self, scan: BlissScanType) -> str:
        """Raw data."""
        filename = scan.scan_info.get("filename")
        if filename:
            return filename
        return current_session.scan_saving.filename

    def get_output_filename(self, scan: BlissScanType) -> str:
        """Processed data."""
        filename = directories.master_output_filename(scan)
        base, ext = os.path.splitext(filename)
        return f"{base}{self.output_suffix or ''}{ext}"

    def get_workflow(self, scan: BlissScanType) -> Optional[str]:
        """Get the workflow to execute for the scan and ensure it is located
        in the proposal directory for user reference and worker accessibility.
        """
        src_file = self._get_workflow(scan)
        if src_file is None:
            return

        dataset_filename = self.get_filename(scan)
        workflow_directory = self._get_workflows_dir(dataset_filename)
        dst_file = os.path.join(workflow_directory, os.path.basename(src_file))
        if src_file != dst_file:
            self._set_workflow(scan, dst_file)

        if not os.path.exists(dst_file):
            os.makedirs(workflow_directory, exist_ok=True)
            shutil.copyfile(src_file, dst_file)

        return dst_file

    def _get_workflow(self, scan: BlissScanType) -> Optional[str]:
        return resource_filename("id16b", "id16b_xeol.json")

    def _set_workflow(self, scan: BlissScanType, filename) -> None:
        self.workflow = filename

    def _get_workflows_dir(self, dataset_filename: str) -> str:
        """Directory where to save the un-parameterized workflow."""
        return directories.get_workflows_dir(dataset_filename)

    def _get_workflow_destination(self, scan: BlissScanType) -> str:
        """Directory where to save the workflow parameterized for each scan."""
        convert_destination = directories.workflow_destination(scan)
        dirname, basename = os.path.split(convert_destination)
        base, ext = os.path.splitext(basename)
        return os.path.join(
            dirname, "workflows", f"{base}{self.output_suffix or ''}{ext}"
        )

    def _get_workflow_upload_parameters(self, scan: BlissScanType) -> dict:
        raw_directory = os.path.dirname(self.get_filename(scan))
        processed_directory = directories.scan_processed_directory(scan)
        scan_saving = current_session.scan_saving
        metadata = {"Sample_name": scan_saving.dataset["Sample_name"]}

        return {
            "beamline": scan_saving.beamline,
            "proposal": scan_saving.proposal_name,
            "dataset": "xeol_fit",
            "path": processed_directory,
            "raw": [raw_directory],
            "metadata": metadata,
        }

    def get_inputs(self, scan: BlissScanType) -> List[dict]:
        if not self.configuration_path:
            raise ValueError(
                "Configuration path is required. Please set it before start"
            )

        inputs = []

        # Raw data parameters

        # Parameters used to read data from HDF5
        input_filename = self.get_filename(scan)
        scan_nb = scan.scan_info.get("scan_nb")

        # Parameters used to read data from Redis
        scan_id = scan._scan_data.key

        for task_identifier in ["ReadStack", "SaveXeolH5"]:
            inputs += [
                {
                    "name": "stack_path",
                    "value": input_filename,
                    "task_identifier": task_identifier,
                },
                {
                    "name": "scan_number",
                    "value": scan_nb,
                    "task_identifier": task_identifier,
                },
                {
                    "name": "key",
                    "value": scan_id,
                    "task_identifier": task_identifier,
                },
            ]

        # Optional read parameters
        if self.main_counter:
            inputs.append(
                {
                    "name": "counter",
                    "value": self.main_counter,
                    "task_identifier": "ReadStack",
                }
            )

        if self.normalization_counter:
            inputs.append(
                {
                    "name": "norm_counter",
                    "value": self.normalization_counter,
                    "task_identifier": "ReadStack",
                }
            )

        if self.threshold is not None:
            inputs.append(
                {
                    "name": "threshold",
                    "value": self.threshold,
                    "task_identifier": "ReadStack",
                }
            )

        # Spectral energy calibration
        calibration_x = self.calibration_x
        if calibration_x is None:
            calibration_x = scan.scan_info.get("instrument", dict()).get("calibration")
        if calibration_x:
            inputs += [
                {
                    "name": "calibration_x",
                    "value": calibration_x,
                    "task_identifier": "ReadStack",
                },
                {
                    "name": "calibration_x",
                    "value": calibration_x,
                    "task_identifier": "SaveXeolH5",
                },
            ]

        # Spectral intensity calibration
        if self.calibration_y:
            inputs += [
                {
                    "name": "calibration_y",
                    "value": self.calibration_y,
                    "task_identifier": "ReadStack",
                },
                {
                    "name": "calibration_y",
                    "value": self.calibration_y,
                    "task_identifier": "SaveXeolH5",
                },
            ]

        # Fit parameters
        inputs.append(
            {
                "name": "config_path",
                "value": self.configuration_path,
                "task_identifier": "ReadCorrectConfig",
            }
        )

        # Save parameters
        output_filename = self.get_output_filename(scan)
        inputs += [
            {
                "name": "save_path",
                "value": output_filename,
                "task_identifier": "SaveXeolH5",
            },
        ]

        return inputs



"""User API for HDF5 conversion on the Bliss repl"""

import logging
import os
import shutil
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from esrf_pathlib import ESRFPath
from ewoksjob.client import submit
from ewoksjob.client.futures import FutureInterface

from ..bliss_globals import current_session
from ..import_utils import unavailable_module
from ..persistent.parameters import ParameterInfo
from ..processor import BaseProcessor
from ..processor import BlissScanType
from ..resources import resource_filename

try:
    import gevent
except ImportError as ex:
    gevent = unavailable_module(ex)


logger = logging.getLogger(__name__)


def _optional_float(value):
    if value is None:
        return None
    return float(value)


def _optional_list_of_strings(value):
    if value is None:
        return None
    if isinstance(value, str):
        return [value]
    return list(map(str, value))


class Bm08Hdf5ToXdiConverter(
    BaseProcessor,
    parameters=[
        ParameterInfo(
            "workflow",
            category="workflows",
            validator=str,
            doc="Workflows to submit for every Kscan.",
        ),
        ParameterInfo(
            "queue",
            category="workflows",
            validator=str,
            doc="Ewoks queue to submit the jobs to.",
        ),
        ParameterInfo(
            "mono_counter",
            category="parameters",
            validator=str,
            doc="Monochromator energy or theta counter name.",
        ),
        ParameterInfo(
            "crystal_motor",
            category="parameters",
            validator=str,
            doc="Motor that selects the monochromator crystal.",
        ),
        ParameterInfo(
            "optional_counters",
            category="parameters",
            validator=_optional_list_of_strings,
            doc="Other counter name to be saved in XDI.",
        ),
        ParameterInfo(
            "optional_mca_counters",
            category="parameters",
            validator=_optional_list_of_strings,
            doc="MCA counter names like ROI names to be saved in XDI.",
        ),
        ParameterInfo(
            "livetime_normalization",
            category="parameters",
            validator=_optional_float,
            doc="Live-time normalization in seconds for the MCA counters.\n"
            " None: no normalization\n"
            " `<=0` the median of the elapsed per point",
        ),
        ParameterInfo(
            "retry_timeout",
            category="data access",
            validator=_optional_float,
            doc="Timeout for HDF5 reading.\n None: wait forever",
        ),
        ParameterInfo(
            "mono_edge_theoretical",
            category="calibration",
            validator=_optional_float,
            doc="The theoretical edge position in 'mono_counter' units.",
        ),
        ParameterInfo(
            "mono_edge_experimental",
            category="calibration",
            validator=_optional_float,
            doc="The experimental edge position in 'mono_counter' units.",
        ),
    ],
):
    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
    ) -> None:
        if defaults is None:
            defaults = dict()
        defaults.setdefault("trigger_at", "END")
        defaults.setdefault("queue", "online")
        defaults.setdefault("mono_counter", "mono_enc")
        defaults.setdefault("crystal_motor", "c_sel")
        defaults.setdefault(
            "optional_counters",
            [
                "I0_eh1",
                "I1_eh1",
                "IX_eh1",
                "I0_eh2",
                "I1_eh2",
                "IX_eh2",
                "IR_eh2",
                "volt1",
                "volt2",
            ],
        )
        defaults.setdefault("optional_mca_counters", [])
        defaults.setdefault("workflow", resource_filename("bm08", "convert.json"))
        defaults.setdefault("retry_timeout", 60)
        defaults.setdefault("livetime_normalization", -1)
        super().__init__(config=config, defaults=defaults)

        # For integration tests
        self._future = None

    def on_new_scan_metadata(self, scan: BlissScanType) -> None:
        if not self.scan_requires_processing(scan):
            return
        workflow = self._get_workflow(scan)
        if not workflow:
            return
        kwargs = self.get_submit_arguments(scan)
        future = submit(args=(workflow,), kwargs=kwargs, queue=self.queue)
        self._future = future

        retry_timeout = self.retry_timeout
        if retry_timeout is not None:
            retry_timeout += 3
        _ = gevent.spawn(_print_result, future, self.retry_timeout)

    def _trigger_workflow_on_new_scan(self, scan: BlissScanType) -> None:
        self.on_new_scan_metadata(scan)

    def scan_requires_processing(self, scan: BlissScanType) -> bool:
        if not scan.scan_info["save"]:
            return False
        scan_parameters = scan.scan_info.get("scan_parameters", {})
        scan_type = scan_parameters.get("scan_type")
        if scan_type != "Kscan":
            return False
        counter_names = self._get_counter_names(scan)
        if self.mono_counter not in counter_names:
            logger.warning(
                "Do not trigger XDI conversion: %r is not used in the scan",
                self.mono_counter,
            )
            return False
        # TODO: check the motor self.crystal_motor is available
        return True

    def get_submit_arguments(self, scan: BlissScanType) -> dict:
        return {
            "inputs": self.get_inputs(scan),
            "outputs": [{"all": False}],
            "convert_destination": self._get_convert_destination(scan),
            "upload_parameters": self._get_workflow_upload_parameters(scan),
        }

    def get_inputs(self, scan: BlissScanType) -> List[dict]:
        load_identifier = "ReadXasHdf5"
        save_identifier = "SaveXasXdi"
        return [
            {
                "task_identifier": load_identifier,
                "name": "filename",
                "value": scan.scan_info["filename"],
            },
            {
                "task_identifier": load_identifier,
                "name": "entry_name",
                "value": f'{scan.scan_info["scan_nb"]}.1',
            },
            {
                "task_identifier": load_identifier,
                "name": "mono_counter",
                "value": self.mono_counter,
            },
            {
                "task_identifier": load_identifier,
                "name": "crystal_motor",
                "value": self.crystal_motor,
            },
            {
                "task_identifier": load_identifier,
                "name": "optional_counters",
                "value": self._get_optional_counters(scan),
            },
            {
                "task_identifier": load_identifier,
                "name": "optional_mca_counters",
                "value": self.optional_mca_counters,
            },
            {
                "task_identifier": load_identifier,
                "name": "livetime_normalization",
                "value": self.livetime_normalization,
            },
            {
                "task_identifier": load_identifier,
                "name": "mono_edge_theoretical",
                "value": self.mono_edge_theoretical,
            },
            {
                "task_identifier": load_identifier,
                "name": "mono_edge_experimental",
                "value": self.mono_edge_experimental,
            },
            {
                "task_identifier": load_identifier,
                "name": "retry_timeout",
                "value": self.retry_timeout,
            },
            {
                "task_identifier": save_identifier,
                "name": "filename",
                "value": self._get_output_filename(scan),
            },
        ]

    def _get_workflow(self, scan: BlissScanType) -> Optional[str]:
        """Get the workflow to execute for the scan and ensure it is located
        in the proposal directory for user reference and worker accessibility.
        """
        src_file = self.workflow
        if src_file is None:
            return
        if not os.path.isfile(src_file):
            src_file = resource_filename("bm08", "xdi_convert.json")

        filename = ESRFPath(scan.scan_info["filename"])
        dst_file = filename.scripts_path / "workflows" / "xdi_convert.json"
        str_dst_file = str(dst_file)

        if src_file != str_dst_file:
            self.workflow = str_dst_file

        if not dst_file.exists():
            dst_file.parent.mkdir(parents=True, exist_ok=True)
            shutil.copyfile(src_file, str_dst_file)

        return str_dst_file

    def _get_counter_names(self, scan: BlissScanType) -> List[str]:
        """Names in the measurement group."""
        counter_names = []
        for full_name in scan.scan_info.get("channels", dict()):
            _, _, name = full_name.rpartition(":")
            counter_names.append(name)
        return counter_names

    def _get_optional_counters(self, scan: BlissScanType) -> List[str]:
        """Names in the measurement group."""
        if not self.optional_counters:
            return []
        counter_names = self._get_counter_names(scan)
        return [s for s in counter_names if s in self.optional_counters]

    def _get_convert_destination(self, scan: BlissScanType) -> str:
        filename = ESRFPath(scan.scan_info["filename"])
        output_directory = filename.processed_dataset_path
        output_stem = f"{filename.stem}_scan{scan.scan_info['scan_nb']}"
        return str(output_directory / "workflows" / f"{output_stem}.json")

    def _get_output_filename(self, scan: BlissScanType) -> str:
        filename = ESRFPath(scan.scan_info["filename"])
        output_directory = filename.processed_dataset_path
        output_stem = scan.scan_info["title"].replace(" ", "_")
        return str(output_directory / f"{output_stem}.xdi")

    def _get_workflow_upload_parameters(self, scan: BlissScanType) -> dict:
        raw_directory = ESRFPath(scan.scan_info["filename"]).raw_dataset_path
        processed_directory = raw_directory.processed_dataset_path
        scan_saving = current_session.scan_saving
        metadata = {"Sample_name": scan_saving.dataset["Sample_name"]}
        return {
            "beamline": scan_saving.beamline,
            "proposal": scan_saving.proposal_name,
            "dataset": "integrate",
            "path": str(processed_directory),
            "raw": [str(raw_directory)],
            "metadata": metadata,
        }


def _print_result(future: FutureInterface, timeout: Optional[float]) -> None:
    try:
        output_filename = future.result(timeout)["output_filename"]
        print(f"XDI file saved: {output_filename}")
    except Exception as ex:
        logger.warning("XDI conversion failed: %s", ex)
